#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Power Plant IDAES heat exchanger model.

The boiler heat exchanger model consist of a cross flow shell and tube hx
that can be used for any of the boiler components, such as economizer,
reheater or superheaters (primary, secondary, etc).
The model includes shell and tube rigorous heat transfer calculations and
pressure drop calculations for shell side. Note that this model assumes no
phase transitions (if user requires phase transitions, they need a general
model)

The main config arguments:
    - delta T method: counter-current or co-current
    - tube_arrangement: in-line or staggered
    - has radiation: True if model is used as a reheater or superheater unit
        Gas emissivity calculated (Gas temperature above 700 K)

General assumptions:
    - SI units (consistent with prop pack)
    - heat transfer calc U = f(Nu, Re, Pr)
    - Pressure drop tube and shell side (friction factor calc.)

"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

__author__ = "Boiler subsystem team (J Ma, M Zamarripa)"
__version__ = "1.0.0"

# Import Python libraries
from enum import Enum

# Import Pyomo libraries
from pyomo.common.config import ConfigValue, In

# Additional import for the unit operation
from pyomo.environ import (
    value,
    Var,
    Param,
    exp,
    sqrt,
    log,
    PositiveReals,
    NonNegativeReals,
    Reference,
    units as pyunits,
)

# Import IDAES cores
from idaes.core import declare_process_block_class

from idaes.core.util.misc import add_object_reference
from idaes.core.util.constants import Constants as c
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.exceptions import ConfigurationError

# These are convenience imports of methods from the base HX model
# pylint: disable=W0611
from idaes.models.unit_models.heat_exchanger import (
    HeatExchangerData,
    delta_temperature_lmtd_callback,
    delta_temperature_lmtd2_callback,
    delta_temperature_lmtd3_callback,
    delta_temperature_amtd_callback,
    delta_temperature_underwood_callback,
    HeatExchangerFlowPattern,
)


# Set up logger
_log = idaeslog.getLogger(__name__)


class TubeArrangement(Enum):
    inLine = 0
    staggered = 1


@declare_process_block_class("BoilerHeatExchanger")
class BoilerHeatExchangerData(HeatExchangerData):
    CONFIG = HeatExchangerData.CONFIG(implicit=True)

    CONFIG.declare(
        "tube_arrangement",
        ConfigValue(
            default=TubeArrangement.inLine,
            domain=In(TubeArrangement),
            description="tube configuration",
            doc="Tube arrangement could be in-line and staggered",
        ),
    )
    CONFIG.declare(
        "cold_side_water_phase",
        ConfigValue(
            default="Liq",
            domain=In(["Liq", "Vap"]),
            description="side 1 water phase",
            doc="Define water phase for property calls",
        ),
    )
    CONFIG.declare(
        "has_radiation",
        ConfigValue(
            default=False,
            domain=In([False, True]),
            description="Has hot side gas radiation",
            doc="Define if hot side gas radiation is to be considered",
        ),
    )

    def _process_config(self):
        """
        Check for boiler specific config arguments.
        """
        config = self.config

        if config.flow_pattern == HeatExchangerFlowPattern.crossflow:
            raise ConfigurationError("Boiler heat exchanger does not support crossflow")

    def build(self):
        """
        Build method for Boiler heat exchanger model

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super().build()
        self._process_config()
        self.deltaT_1 = Reference(self.delta_temperature_in)
        self.deltaT_2 = Reference(self.delta_temperature_out)

        self._set_geometry()
        self.cold_side_fluid_phase = self.config.cold_side_water_phase
        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary, and link to holdup volume

        Args:
            None

        Returns:
            None
        """
        # Elevation difference (outlet - inlet) for static pressure calculation
        self.delta_elevation = Var(
            initialize=0,
            within=NonNegativeReals,
            doc="Elevation increase used for static pressure calculation",
            units=pyunits.m,
        )

        # Number of tube columns in the cross section plane
        # perpendicular to shell side fluid flow (y direction)
        self.tube_ncol = Var(
            initialize=10.0, within=PositiveReals, doc="Number of tube columns"
        )

        # Number of tube rows in the direction of shell side
        # fluid flow (x direction)
        self.tube_nrow = Var(
            initialize=10.0, within=PositiveReals, doc="Number of tube rows"
        )

        # Number of inlet tube rows
        self.nrow_inlet = Var(
            initialize=1, within=PositiveReals, doc="Number of inlet tube rows"
        )

        # Length of a tube in z direction for each path
        self.tube_length = Var(
            initialize=5.0, within=PositiveReals, doc="Tube length", units=pyunits.m
        )

        # Inner diameter of tubes
        self.tube_di = Var(
            initialize=0.05,
            within=PositiveReals,
            doc="Inner diameter of tube",
            units=pyunits.m,
        )

        # Thickness of tube
        self.tube_thickness = Var(
            initialize=0.005,
            within=PositiveReals,
            doc="Tube thickness",
            units=pyunits.m,
        )

        # Pitch of tubes between two neighboring columns (in y direction).
        # Always greater than tube outside diameter
        self.pitch_y = Var(
            initialize=0.1,
            within=PositiveReals,
            doc="Pitch between two neighboring columns",
            units=pyunits.m,
        )

        # Pitch of tubes between two neighboring rows (in x direction).
        # Always greater than tube outside diameter
        self.pitch_x = Var(
            initialize=0.1,
            within=PositiveReals,
            doc="Pitch between two neighboring rows",
            units=pyunits.m,
        )

        # Tube outside diameter
        @self.Expression(doc="Outside diameter of tube")
        def do_tube(b):
            return b.tube_di + b.tube_thickness * 2.0

        if self.config.has_radiation is True:
            # Mean beam length for radiation
            @self.Expression(doc="Mean beam length")
            def mbl(b):
                return 3.6 * (
                    b.pitch_x * b.pitch_y / c.pi / b.do_tube - b.do_tube / 4.0
                )

            # Mean beam length for radiation divided by sqrt(2)
            @self.Expression(doc="Mean beam length")
            def mbl_div2(b):
                return b.mbl / sqrt(2.0)

            # Mean beam length for radiation multiplied by sqrt(2)
            @self.Expression(doc="Mean beam length")
            def mbl_mul2(b):
                return b.mbl * sqrt(2.0)

        # Number of 180 degree bends for the tube
        @self.Expression(doc="Nbend_tube")
        def nbend_tube(b):
            return b.tube_nrow / b.nrow_inlet

        # Total flow area on tube side
        @self.Expression(doc="Total flow area on tube side")
        def area_flow_tube(b):
            return 0.25 * c.pi * b.tube_di**2.0 * b.tube_ncol * b.nrow_inlet

        # Total flow area on shell side
        @self.Expression(doc="Total flow area on shell side")
        def area_flow_shell(b):
            return b.tube_length * (b.pitch_y - b.do_tube) * b.tube_ncol

        # Total heat transfer area based on outside diameter
        @self.Constraint(doc="Total heat transfer area based on tube outside diameter")
        def area_eqn(b):
            return (
                b.area == c.pi * b.do_tube * b.tube_length * b.tube_ncol * b.tube_nrow
            )

        # Ratio of pitch_x/do_tube
        @self.Expression(
            doc="Ratio of pitch in x " "direction to tube outside diameter"
        )
        def pitch_x_to_do(b):
            return b.pitch_x / b.do_tube

        # Ratio of pitch_y/do_tube
        @self.Expression(
            doc="Ratio of pitch in y " "direction to tube outside diameter"
        )
        def pitch_y_to_do(b):
            return b.pitch_y / b.do_tube

        if self.config.has_holdup is True:
            add_object_reference(self, "volume_cold_side", self.cold_side.volume)
            add_object_reference(self, "volume_hot_side", self.hot_side.volume)

            # Total tube side valume
            def volume_cold_side_eqn(b):
                return b.volume_cold_side == (
                    0.25
                    * c.pi
                    * b.tube_di**2.0
                    * b.tube_length
                    * b.tube_ncol
                    * b.tube_nrow
                )

            self.Constraint(doc="Total tube side volume", rule=volume_cold_side_eqn)

            def volume_hot_side_eqn(b):
                return (
                    b.volume_hot_side
                    == b.tube_ncol * b.pitch_y * b.tube_length * b.tube_nrow * b.pitch_x
                    - 0.25
                    * c.pi
                    * b.do_tube**2.0
                    * b.tube_length
                    * b.tube_ncol
                    * b.tube_nrow
                )

            # Total shell side volume
            self.Constraint(doc="Total shell side volume", rule=volume_hot_side_eqn)

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.

        Args:
            None

        Returns:
            None
        """
        # Set references to balance terms at unit level
        add_object_reference(self, "heat_duty", self.cold_side.heat)
        if self.config.cold_side.has_pressure_change is True:
            add_object_reference(self, "deltaP_tube", self.cold_side.deltaP)
        if self.config.hot_side.has_pressure_change is True:
            add_object_reference(self, "deltaP_shell", self.hot_side.deltaP)

        # Performance parameters and variables
        # Wall thermal conductivity
        self.therm_cond_wall = Param(
            initialize=43.0,
            within=PositiveReals,
            doc="Thermal conductivity of the wall",
            units=pyunits.W / pyunits.m / pyunits.K,
        )

        # Loss coefficient for a 180 degree bend (u-turn),
        # usually related to radius to inside diameter ratio
        self.k_loss_uturn = Param(
            initialize=0.5,
            within=PositiveReals,
            mutable=True,
            doc="Loss coefficient of a tube u-turn",
        )

        # Heat transfer resistance due to the fouling on tube side
        # (typical boiler hx)
        self.tube_r_fouling = Param(
            initialize=0.00017,
            within=NonNegativeReals,
            mutable=True,
            doc="Fouling resistance on tube side",
            units=pyunits.K * pyunits.m**2 * pyunits.W**-1,
        )

        # Heat transfer resistance due to the fouling on shell side
        self.shell_r_fouling = Param(
            initialize=0.0008,
            within=NonNegativeReals,
            mutable=True,
            doc="Fouling resistance on tube side",
            units=pyunits.K * pyunits.m**2 * pyunits.W**-1,
        )

        # Correction factor for overall heat transfer coefficient
        self.fcorrection_htc = Var(
            initialize=1.0, within=NonNegativeReals, doc="Correction factor for HTC"
        )

        # Correction factor for tube side pressure drop due to friction
        self.fcorrection_dp_tube = Var(
            initialize=1.0, doc="Correction factor for tube side pressure drop"
        )

        # Correction factor for shell side pressure drop due to friction
        self.fcorrection_dp_shell = Var(
            initialize=1.0, doc="Correction factor for shell side pressure drop"
        )

        if self.config.has_radiation is True:
            # Shell side wall emissivity, converted from parameter to variable
            self.emissivity_wall = Var(initialize=0.7, doc="Shell side wall emissivity")
            # Gas emissivity at mbl
            self.gas_emissivity = Var(
                self.flowsheet().time,
                initialize=0.5,
                doc="Emissivity at given mean beam length",
            )

            # Gas emissivity at mbl/sqrt(2)
            self.gas_emissivity_div2 = Var(
                self.flowsheet().time,
                initialize=0.4,
                doc="Emissivity at mean beam length divided by sqrt of 2",
            )

            # Gas emissivity at mbl*sqrt(2)
            self.gas_emissivity_mul2 = Var(
                self.flowsheet().time,
                initialize=0.6,
                doc="Emissivity at mean beam length multiplied by sqrt of 2",
            )

            # Gray fraction of gas in entire spectrum
            self.gas_gray_fraction = Var(
                self.flowsheet().time,
                initialize=0.5,
                doc="Gray fraction of gas in entire spectrum",
            )

            # Gas-surface radiation exchange factor for shell side wall
            self.frad_gas_shell = Var(
                self.flowsheet().time,
                initialize=0.5,
                doc="Gas-surface radiation exchange " "factor for shell side wall",
            )

            # Shell side equivalent convective heat transfer coefficient
            # due to radiation
            self.hconv_shell_rad = Var(
                self.flowsheet().time,
                initialize=100.0,
                doc="Shell convective heat transfer coefficient due to radiation",
                units=pyunits.W / pyunits.m**2 / pyunits.K,
            )

        # Tube side convective heat transfer coefficient
        self.hconv_tube = Var(
            self.flowsheet().time,
            initialize=100.0,
            doc="Tube side convective heat transfer coefficient",
            units=pyunits.W / pyunits.m**2 / pyunits.K,
        )

        # Shell side convective heat transfer coefficient due to convection
        self.hconv_shell_conv = Var(
            self.flowsheet().time,
            initialize=100.0,
            doc="Shell side convective heat transfer coefficient due to convection",
            units=pyunits.W / pyunits.m**2 / pyunits.K,
        )

        # Total shell side convective heat transfer coefficient
        # including convection and radiation
        self.hconv_shell_total = Var(
            self.flowsheet().time,
            initialize=150.0,
            doc="Total shell side convective heat transfer coefficient",
            units=pyunits.W / pyunits.m**2 / pyunits.K,
        )

        # Heat conduction resistance of tube wall
        self.rcond_wall = Var(
            initialize=1.0,
            doc="Heat conduction resistance of wall",
            units=pyunits.m**2 * pyunits.K / pyunits.W,
        )

        if self.config.has_radiation is True:
            # Constraints for gas emissivity
            @self.Constraint(self.flowsheet().time, doc="Gas emissivity")
            def gas_emissivity_eqn(b, t):
                # This is a surrogate model, so need to do units manually
                X1 = (
                    (
                        b.hot_side.properties_in[t].temperature
                        + b.hot_side.properties_out[t].temperature
                    )
                    / 2
                    / pyunits.K
                )
                X2 = b.mbl / pyunits.m
                X3 = b.hot_side.properties_in[t].pressure / pyunits.Pa
                X4 = b.hot_side.properties_in[t].mole_frac_comp["CO2"]
                X5 = b.hot_side.properties_in[t].mole_frac_comp["H2O"]
                X6 = b.hot_side.properties_in[t].mole_frac_comp["O2"]

                # Surrogate model fitted using rigorous calc. - 500 samples
                # Wide operating range:
                #      X1: 700 – 1500    (Gas Temperature)
                #      X2: 0.2 – 1       (Mean beam length)
                #      X3: 79000-102000  (pressure in Pa)
                #      X4: 0.12-0.16     (mol frac CO2)
                #      X5: 0.075-0.15    (mol frac H2O)
                #      X6: 0.01-0.07     (mol frac O2)

                return b.gas_emissivity[t] == (
                    -0.116916606892e-003 * X1
                    - 0.29111124038936179309056e-001 * X2
                    + 0.50509651230704191577346e-006 * X3
                    + 1.1844222822155641150488 * X4
                    - 0.64720757767102773949652e-001 * X5
                    - 0.35853593221454795048064e-001 * X6
                    + 0.12227919099126832724878 * log(X1)
                    + 0.45102118316418124410738e-001 * log(X2)
                    + 0.33111863480179408447679e-001 * log(X3)
                    + 0.17674928397780117345084e-001 * log(X5)
                    - 0.12541139396423576016226e-001 * exp(X2)
                    - 0.90251708836308952577099 * exp(X4)
                    + 0.32447078857791738538963e-002 * X2**2
                    - 0.31332075610864829615706e-004 * X1 * X2
                    - 0.54639645449809960433102e-009 * X1 * X3
                    - 0.19721467902854980460033e-003 * X1 * X5
                    + 0.45275517692290622763507e-004 * X1 * X6
                    + 0.75458754990630776904396e-006 * X2 * X3
                    + 0.39691751689931338564765e-001 * X2 * X4
                    + 0.73169514231974708273754 * X2 * X5
                    - 0.35852614507684822664491e-001 * X2 * X6
                    + 0.39743672195685803976177e-005 * X3 * X5
                    + 0.58802879141883679897383e-008 * (X1 * X2) ** 2
                    - 1.2994610452829884472692 * (X2 * X5) ** 2
                )

            # Constraints for gas emissivity at mbl/sqrt(2)
            @self.Constraint(
                self.flowsheet().time, doc="Gas emissivity at a lower mean beam length"
            )
            def gas_emissivity_div2_eqn(b, t):
                # This is a surrogate model, so need to do units manually
                X1 = (
                    (
                        b.hot_side.properties_in[t].temperature
                        + b.hot_side.properties_out[t].temperature
                    )
                    / 2
                    / pyunits.K
                )
                X2 = b.mbl_div2 / pyunits.m
                X3 = b.hot_side.properties_in[t].pressure / pyunits.Pa
                X4 = b.hot_side.properties_in[t].mole_frac_comp["CO2"]
                X5 = b.hot_side.properties_in[t].mole_frac_comp["H2O"]
                X6 = b.hot_side.properties_in[t].mole_frac_comp["O2"]

                # Surrogate model fitted using rigorous calc. - 500 samples
                # Wide operating range:
                #       X1: 700 – 1500    (Gas Temperature)
                #       X2: 0.2 – 1       (Mean beam length)
                #       X3: 79000-102000  (pressure in Pa)
                #       X4: 0.12-0.16     (mol frac CO2)
                #       X5: 0.075-0.15    (mol frac H2O)
                #       X6: 0.01-0.07     (mol frac O2)
                return b.gas_emissivity_div2[t] == (
                    -0.116916606892e-003 * X1
                    - 0.29111124038936179309056e-001 * X2
                    + 0.50509651230704191577346e-006 * X3
                    + 1.1844222822155641150488 * X4
                    - 0.64720757767102773949652e-001 * X5
                    - 0.35853593221454795048064e-001 * X6
                    + 0.12227919099126832724878 * log(X1)
                    + 0.45102118316418124410738e-001 * log(X2)
                    + 0.33111863480179408447679e-001 * log(X3)
                    + 0.17674928397780117345084e-001 * log(X5)
                    - 0.12541139396423576016226e-001 * exp(X2)
                    - 0.90251708836308952577099 * exp(X4)
                    + 0.32447078857791738538963e-002 * X2**2
                    - 0.31332075610864829615706e-004 * X1 * X2
                    - 0.54639645449809960433102e-009 * X1 * X3
                    - 0.19721467902854980460033e-003 * X1 * X5
                    + 0.45275517692290622763507e-004 * X1 * X6
                    + 0.75458754990630776904396e-006 * X2 * X3
                    + 0.39691751689931338564765e-001 * X2 * X4
                    + 0.73169514231974708273754 * X2 * X5
                    - 0.35852614507684822664491e-001 * X2 * X6
                    + 0.39743672195685803976177e-005 * X3 * X5
                    + 0.58802879141883679897383e-008 * (X1 * X2) ** 2
                    - 1.2994610452829884472692 * (X2 * X5) ** 2
                )

            # Constraints for gas emissivity at mbl*sqrt(2)
            @self.Constraint(
                self.flowsheet().time, doc="Gas emissivity at a higher mean beam length"
            )
            def gas_emissivity_mul2_eqn(b, t):
                # This is a surrogate model, so need to do units manually
                X1 = (
                    (
                        b.hot_side.properties_in[t].temperature
                        + b.hot_side.properties_out[t].temperature
                    )
                    / 2
                    / pyunits.K
                )
                X2 = b.mbl_mul2 / pyunits.m
                X3 = b.hot_side.properties_in[t].pressure / pyunits.Pa
                X4 = b.hot_side.properties_in[t].mole_frac_comp["CO2"]
                X5 = b.hot_side.properties_in[t].mole_frac_comp["H2O"]
                X6 = b.hot_side.properties_in[t].mole_frac_comp["O2"]

                # Surrogate model fitted using rigorous calc. 500 samples
                # Wide operating range:
                #       X1: 700 – 1500    (Gas Temperature)
                #       X2: 0.2 – 1       (Mean beam length)
                #       X3: 79000-102000  (pressure in Pa)
                #       X4: 0.12-0.16     (mol frac CO2)
                #       X5: 0.075-0.15    (mol frac H2O)
                #       X6: 0.01-0.07     (mol frac O2)
                return b.gas_emissivity_mul2[t] == (
                    -0.116916606892e-003 * X1
                    - 0.29111124038936179309056e-001 * X2
                    + 0.50509651230704191577346e-006 * X3
                    + 1.1844222822155641150488 * X4
                    - 0.64720757767102773949652e-001 * X5
                    - 0.35853593221454795048064e-001 * X6
                    + 0.12227919099126832724878 * log(X1)
                    + 0.45102118316418124410738e-001 * log(X2)
                    + 0.33111863480179408447679e-001 * log(X3)
                    + 0.17674928397780117345084e-001 * log(X5)
                    - 0.12541139396423576016226e-001 * exp(X2)
                    - 0.90251708836308952577099 * exp(X4)
                    + 0.32447078857791738538963e-002 * X2**2
                    - 0.31332075610864829615706e-004 * X1 * X2
                    - 0.54639645449809960433102e-009 * X1 * X3
                    - 0.19721467902854980460033e-003 * X1 * X5
                    + 0.45275517692290622763507e-004 * X1 * X6
                    + 0.75458754990630776904396e-006 * X2 * X3
                    + 0.39691751689931338564765e-001 * X2 * X4
                    + 0.73169514231974708273754 * X2 * X5
                    - 0.35852614507684822664491e-001 * X2 * X6
                    + 0.39743672195685803976177e-005 * X3 * X5
                    + 0.58802879141883679897383e-008 * (X1 * X2) ** 2
                    - 1.2994610452829884472692 * (X2 * X5) ** 2
                )

            # fraction of gray gas spectrum
            @self.Constraint(self.flowsheet().time, doc="Fraction of gray gas spectrum")
            def gas_gray_fraction_eqn(b, t):
                return (
                    b.gas_gray_fraction[t]
                    * (2 * b.gas_emissivity_div2[t] - b.gas_emissivity_mul2[t])
                    == b.gas_emissivity_div2[t] ** 2
                )

            # gas-surface radiation exchange factor
            # between gas and shell side wall
            @self.Constraint(
                self.flowsheet().time,
                doc="Gas-surface radiation exchange "
                "factor between gas and shell side wall",
            )
            def frad_gas_shell_eqn(b, t):
                return (
                    b.frad_gas_shell[t]
                    * (
                        (1 / b.emissivity_wall - 1) * b.gas_emissivity[t]
                        + b.gas_gray_fraction[t]
                    )
                    == b.gas_gray_fraction[t] * b.gas_emissivity[t]
                )

            # equivalent convective heat transfer coefficient due to radiation
            @self.Constraint(
                self.flowsheet().time,
                doc="Equivalent convective heat transfer "
                "coefficient due to radiation",
            )
            def hconv_shell_rad_eqn(b, t):
                return b.hconv_shell_rad[t] == c.stefan_constant * b.frad_gas_shell[
                    t
                ] * (
                    (
                        b.hot_side.properties_in[t].temperature
                        + b.hot_side.properties_out[t].temperature
                    )
                    / 2
                    + b.cold_side.properties_in[t].temperature
                ) * (
                    (
                        (
                            b.hot_side.properties_in[t].temperature
                            + b.hot_side.properties_out[t].temperature
                        )
                        / 2
                    )
                    ** 2
                    + b.cold_side.properties_in[t].temperature ** 2
                )

        # Tube side heat transfer coefficient and pressure drop
        # -----------------------------------------------------
        # Velocity on tube side
        self.v_tube = Var(
            self.flowsheet().time,
            initialize=1.0,
            doc="Velocity on tube side",
            units=pyunits.m / pyunits.s,
        )

        # Reynalds number on tube side
        self.N_Re_tube = Var(
            self.flowsheet().time,
            initialize=10000.0,
            doc="Reynolds number on tube side",
        )
        if self.config.cold_side.has_pressure_change is True:
            # Friction factor on tube side
            self.friction_factor_tube = Var(
                self.flowsheet().time,
                initialize=1.0,
                doc="Friction factor on tube side",
            )

            # Pressure drop due to friction on tube side
            self.deltaP_tube_friction = Var(
                self.flowsheet().time,
                initialize=-10.0,
                doc="Pressure drop due to friction on tube side",
                units=pyunits.Pa,
            )

            # Pressure drop due to 180 degree turn on tube side
            self.deltaP_tube_uturn = Var(
                self.flowsheet().time,
                initialize=-10.0,
                doc="Pressure drop due to u-turn on tube side",
                units=pyunits.Pa,
            )

        # Prandtl number on tube side
        self.N_Pr_tube = Var(
            self.flowsheet().time, initialize=1, doc="Prandtl number on tube side"
        )

        # Nusselt number on tube side
        self.N_Nu_tube = Var(
            self.flowsheet().time, initialize=1, doc="Nusselts number on tube side"
        )

        # Velocity equation
        @self.Constraint(self.flowsheet().time, doc="Tube side velocity equation")
        def v_tube_eqn(b, t):
            return (
                b.v_tube[t]
                * b.area_flow_tube
                * b.cold_side.properties_in[t].dens_mol_phase[
                    self.cold_side_fluid_phase
                ]
                == b.cold_side.properties_in[t].flow_mol
            )

        # Reynolds number
        @self.Constraint(
            self.flowsheet().time, doc="Reynolds number equation on tube side"
        )
        def N_Re_tube_eqn(b, t):
            return (
                b.N_Re_tube[t]
                * b.cold_side.properties_in[t].visc_d_phase[self.cold_side_fluid_phase]
                == b.tube_di
                * b.v_tube[t]
                * b.cold_side.properties_in[t].dens_mass_phase[
                    self.cold_side_fluid_phase
                ]
            )

        if self.config.cold_side.has_pressure_change is True:
            # Friction factor
            @self.Constraint(
                self.flowsheet().time, doc="Darcy friction factor on tube side"
            )
            def friction_factor_tube_eqn(b, t):
                return (
                    b.friction_factor_tube[t] * b.N_Re_tube[t] ** 0.25
                    == 0.3164 * b.fcorrection_dp_tube
                )

            # Pressure drop due to friction
            @self.Constraint(
                self.flowsheet().time, doc="Pressure drop due to friction on tube side"
            )
            def deltaP_tube_friction_eqn(b, t):
                return (
                    b.deltaP_tube_friction[t]
                    == -0.5
                    * b.cold_side.properties_in[t].dens_mass_phase[
                        self.cold_side_fluid_phase
                    ]
                    * b.v_tube[t] ** 2
                    * b.friction_factor_tube[t]
                    * b.tube_length
                    * b.tube_nrow
                    / b.tube_di
                    / b.nrow_inlet
                )

            # Pressure drop due to u-turn
            @self.Constraint(
                self.flowsheet().time, doc="Pressure drop due to u-turn on tube side"
            )
            def deltaP_tube_uturn_eqn(b, t):
                return (
                    b.deltaP_tube_uturn[t]
                    == -0.5
                    * b.cold_side.properties_in[t].dens_mass_phase[
                        self.cold_side_fluid_phase
                    ]
                    * b.v_tube[t] ** 2
                    * b.k_loss_uturn
                )

            # Total pressure drop on tube side
            @self.Constraint(
                self.flowsheet().time, doc="Total pressure drop on tube side"
            )
            def deltaP_tube_eqn(b, t):
                return (
                    b.deltaP_tube[t]
                    == b.deltaP_tube_friction[t]
                    + b.deltaP_tube_uturn[t]
                    - b.delta_elevation
                    * c.acceleration_gravity
                    * (
                        b.cold_side.properties_in[t].dens_mass_phase[
                            self.cold_side_fluid_phase
                        ]
                        + b.cold_side.properties_out[t].dens_mass_phase[
                            self.cold_side_fluid_phase
                        ]
                    )
                    / 2.0
                )

        # Prandtl number
        @self.Constraint(
            self.flowsheet().time, doc="Prandtl number equation on tube side"
        )
        def N_Pr_tube_eqn(b, t):
            return (
                b.N_Pr_tube[t]
                * b.cold_side.properties_in[t].therm_cond_phase[
                    self.cold_side_fluid_phase
                ]
                * b.cold_side.properties_in[t].mw
                == b.cold_side.properties_in[t].cp_mol_phase[self.cold_side_fluid_phase]
                * b.cold_side.properties_in[t].visc_d_phase[self.cold_side_fluid_phase]
            )

        # Nusselts number
        @self.Constraint(
            self.flowsheet().time, doc="Nusselts number equation on tube side"
        )
        def N_Nu_tube_eqn(b, t):
            return (
                b.N_Nu_tube[t]
                == 0.023 * b.N_Re_tube[t] ** 0.8 * abs(b.N_Pr_tube[t]) ** 0.4
            )

        # Heat transfer coefficient
        @self.Constraint(
            self.flowsheet().time,
            doc="Convective heat transfer " "coefficient equation on tube side",
        )
        def hconv_tube_eqn(b, t):
            return (
                b.hconv_tube[t] * self.tube_di / 1000
                == b.N_Nu_tube[t]
                * b.cold_side.properties_in[t].therm_cond_phase[
                    self.cold_side_fluid_phase
                ]
                / 1000
            )

        # Pressure drop and heat transfer coefficient on shell side
        # ----------------------------------------------------------
        # Tube arrangement factor
        if self.config.tube_arrangement == TubeArrangement.inLine:
            self.f_arrangement = Param(
                initialize=0.788, doc="In-line tube arrangement factor"
            )
        elif self.config.tube_arrangement == TubeArrangement.staggered:
            self.f_arrangement = Param(
                initialize=1.0, doc="Staggered tube arrangement factor"
            )
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("tube arrangement type not supported")
        # Velocity on shell side
        self.v_shell = Var(
            self.flowsheet().time,
            initialize=1.0,
            doc="Velocity on shell side",
            units=pyunits.m / pyunits.s,
        )

        # Reynalds number on shell side
        self.N_Re_shell = Var(
            self.flowsheet().time,
            initialize=10000.0,
            doc="Reynolds number on shell side",
        )

        # Friction factor on shell side
        self.friction_factor_shell = Var(
            self.flowsheet().time, initialize=1.0, doc="Friction factor on shell side"
        )

        # Prandtl number on shell side
        self.N_Pr_shell = Var(
            self.flowsheet().time, initialize=1, doc="Prandtl number on shell side"
        )

        # Nusselt number on shell side
        self.N_Nu_shell = Var(
            self.flowsheet().time, initialize=1, doc="Nusselts number on shell side"
        )

        # Velocity equation on shell side
        @self.Constraint(self.flowsheet().time, doc="Velocity on shell side")
        def v_shell_eqn(b, t):
            return b.v_shell[t] * b.hot_side.properties_in[t].dens_mol_phase[
                "Vap"
            ] * b.area_flow_shell == sum(
                b.hot_side.properties_in[t].flow_mol_comp[j]
                for j in b.hot_side.properties_in[t].params.component_list
            )

        # Reynolds number
        @self.Constraint(
            self.flowsheet().time, doc="Reynolds number equation on shell side"
        )
        def N_Re_shell_eqn(b, t):
            return b.N_Re_shell[t] * b.hot_side.properties_in[
                t
            ].visc_d == b.do_tube * b.v_shell[t] * b.hot_side.properties_in[
                t
            ].dens_mol_phase[
                "Vap"
            ] * sum(
                b.hot_side.properties_in[t].mw_comp[c]
                * b.hot_side.properties_in[t].mole_frac_comp[c]
                for c in b.hot_side.properties_in[t].params.component_list
            )

        if self.config.hot_side.has_pressure_change is True:
            # Friction factor on shell side
            if self.config.tube_arrangement == TubeArrangement.inLine:

                @self.Constraint(
                    self.flowsheet().time, doc="In-line friction factor on shell side"
                )
                def friction_factor_shell_eqn(b, t):
                    return (
                        b.friction_factor_shell[t] * b.N_Re_shell[t] ** 0.15
                        == (
                            0.044
                            + 0.08
                            * b.pitch_x_to_do
                            / (b.pitch_y_to_do - 1.0) ** (0.43 + 1.13 / b.pitch_x_to_do)
                        )
                        * b.fcorrection_dp_shell
                    )

            elif self.config.tube_arrangement == TubeArrangement.staggered:

                @self.Constraint(
                    self.flowsheet().time, doc="Staggered friction factor on shell side"
                )
                def friction_factor_shell_eqn(b, t):
                    return (
                        b.friction_factor_shell[t] * b.N_Re_shell[t] ** 0.16
                        == (0.25 + 0.118 / (b.pitch_y_to_do - 1.0) ** 1.08)
                        * b.fcorrection_dp_shell
                    )

            else:
                # PYLINT-TODO
                # pylint: disable-next=broad-exception-raised
                raise Exception("tube arrangement type not supported")

            # Pressure drop on shell side
            @self.Constraint(self.flowsheet().time, doc="Pressure change on shell side")
            def deltaP_shell_eqn(b, t):
                return (
                    b.deltaP_shell[t]
                    == -1.4
                    * b.friction_factor_shell[t]
                    * b.tube_nrow
                    * b.hot_side.properties_in[t].dens_mol_phase["Vap"]
                    * sum(
                        b.hot_side.properties_in[t].mw_comp[c]
                        * b.hot_side.properties_in[t].mole_frac_comp[c]
                        for c in b.hot_side.properties_in[t].params.component_list
                    )
                    * b.v_shell[t] ** 2
                )

        # Prandtl number
        @self.Constraint(
            self.flowsheet().time, doc="Prandtl number equation on shell side"
        )
        def N_Pr_shell_eqn(b, t):
            return (
                b.N_Pr_shell[t]
                * b.hot_side.properties_in[t].therm_cond
                * sum(
                    b.hot_side.properties_in[t].mw_comp[c]
                    * b.hot_side.properties_in[t].mole_frac_comp[c]
                    for c in b.hot_side.properties_in[t].params.component_list
                )
                == b.hot_side.properties_in[t].cp_mol
                * b.hot_side.properties_in[t].visc_d
            )

        # Nusselt number, currently assume Re>300
        @self.Constraint(
            self.flowsheet().time, doc="Nusselts number equation on shell side"
        )
        def N_Nu_shell_eqn(b, t):
            return (
                b.N_Nu_shell[t]
                == b.f_arrangement
                * 0.33
                * b.N_Re_shell[t] ** 0.6
                * b.N_Pr_shell[t] ** 0.333333
            )

        # Convective heat transfer coefficient on shell side due to convection
        @self.Constraint(
            self.flowsheet().time,
            doc="Convective heat transfer coefficient equation"
            "on shell side due to convection",
        )
        def hconv_shell_conv_eqn(b, t):
            return (
                b.hconv_shell_conv[t] * b.do_tube / 1000
                == b.N_Nu_shell[t] * b.hot_side.properties_in[t].therm_cond / 1000
            )

        # Total convective heat transfer coefficient on shell side
        @self.Constraint(
            self.flowsheet().time,
            doc="Total convective heat transfer " "coefficient equation on shell side",
        )
        def hconv_shell_total_eqn(b, t):
            if self.config.has_radiation is True:
                return (
                    b.hconv_shell_total[t]
                    == b.hconv_shell_conv[t] + b.hconv_shell_rad[t]
                )
            else:
                return b.hconv_shell_total[t] == b.hconv_shell_conv[t]

        # Wall conduction heat transfer resistance
        # based on outside surface area
        @self.Constraint(doc="Wall conduction heat transfer resistance")
        def rcond_wall_eqn(b):
            return b.rcond_wall * b.therm_cond_wall == 0.5 * b.do_tube * log(
                b.do_tube / b.tube_di
            )

        # Overall heat transfer coefficient
        @self.Constraint(
            self.flowsheet().time, doc="Wall conduction heat transfer resistance"
        )
        def overall_heat_transfer_coefficient_eqn(b, t):
            return (
                b.overall_heat_transfer_coefficient[t]
                * (
                    b.rcond_wall
                    + b.tube_r_fouling
                    + b.shell_r_fouling
                    + 1.0 / b.hconv_shell_total[t]
                    + b.do_tube / b.hconv_tube[t] / b.tube_di
                )
                == b.fcorrection_htc
            )

    def model_check(blk):
        """
        Model checks for unit - calls model checks for both control volume
        Blocks.

        Args:
            None

        Returns:
            None
        """
        # Run control volume block model checks
        blk.cold_side.model_check()
        blk.hot_side.model_check()

    def initialize_build(
        blk,
        state_args_1=None,
        state_args_2=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General Heat Exchanger initialisation routine.

        Keyword Arguments:
            state_args_1 : a dict of arguments to be passed to the property
                           package(s) for side 1 of the heat exchanger to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            state_args_2 : a dict of arguments to be passed to the property
                           package(s) for side 2 of the heat exchanger to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize inlet property blocks
        flags1 = blk.cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_1
        )

        flags2 = blk.hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_2
        )
        init_log.info("{} Initialisation Step 1 Complete.".format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize temperature differentials
        p1_flags = {}
        p2_flags = {}
        h1_flags = {}
        t2_flags = {}
        for t in blk.flowsheet().time:
            p1_flags[t] = blk.cold_side.properties_out[t].pressure.fixed
            if (
                not blk.cold_side.properties_out[t].pressure.fixed
                and blk.config.cold_side.has_pressure_change
            ):
                blk.cold_side.properties_out[t].pressure.fix(
                    value(blk.cold_side.properties_in[t].pressure)
                )

            p2_flags[t] = blk.hot_side.properties_out[t].pressure.fixed
            if (
                not blk.hot_side.properties_out[t].pressure.fixed
                and blk.config.hot_side.has_pressure_change
            ):
                blk.hot_side.properties_out[t].pressure.fix(
                    value(blk.hot_side.properties_in[t].pressure)
                )

            h1_flags[t] = blk.cold_side.properties_out[t].enth_mol.fixed
            if not blk.cold_side.properties_out[t].enth_mol.fixed:
                blk.cold_side.properties_out[t].enth_mol.fix(
                    value(blk.cold_side.properties_in[t].enth_mol) + 100.0
                )

            t2_flags[t] = blk.hot_side.properties_out[t].temperature.fixed
            if not blk.hot_side.properties_out[t].temperature.fixed:
                blk.hot_side.properties_out[t].temperature.fix(
                    value(blk.hot_side.properties_in[t].temperature) - 5.0
                )
                #                                assuming Delta T min approach
        # Deactivate Constraints
        blk.heat_transfer_equation.deactivate()
        blk.unit_heat_balance.deactivate()
        if blk.config.cold_side.has_pressure_change:
            blk.deltaP_tube_eqn.deactivate()
        if blk.config.hot_side.has_pressure_change:
            blk.deltaP_shell_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Activate energy balance and driving force
        for t in blk.flowsheet().time:
            if not p1_flags[t]:
                blk.cold_side.properties_out[t].pressure.unfix()
            if not p2_flags[t]:
                blk.hot_side.properties_out[t].pressure.unfix()
            if not h1_flags[t]:
                blk.cold_side.properties_out[t].enth_mol.unfix()
            if not t2_flags[t]:
                blk.hot_side.properties_out[t].temperature.unfix()
        blk.heat_transfer_equation.activate()
        blk.unit_heat_balance.activate()

        if blk.config.cold_side.has_pressure_change:
            blk.deltaP_tube_eqn.activate()
        if blk.config.hot_side.has_pressure_change:
            blk.deltaP_shell_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.cold_side.release_state(flags1, outlvl)
        blk.hot_side.release_state(flags2, outlvl)

        init_log.info("{} Initialisation Complete.".format(blk.name))

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # Since this depends on the process size this is another scaling factor
        # the user should always set.
        sf_a = iscale.get_scaling_factor(self.area, default=1e-4, warning=True)

        iscale.constraint_scaling_transform(self.area_eqn, sf_a)

        for t, c in self.v_shell_eqn.items():
            s = iscale.min_scaling_factor(
                self.hot_side.properties_in[t].flow_mol_comp,
                default=0,
                warning=False,
                hint=None,
            )
            if s == 0:
                s = iscale.get_scaling_factor(
                    self.hot_side.properties_in[t].flow_mol, default=1, warning=True
                )
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        for t, c in self.v_tube_eqn.items():
            s = iscale.min_scaling_factor(
                self.cold_side.properties_in[t].flow_mol_comp,
                default=0,
                warning=False,
                hint=None,
            )
            if s == 0:
                s = iscale.get_scaling_factor(
                    self.cold_side.properties_in[t].flow_mol, default=1, warning=True
                )
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        for t, c in self.N_Nu_tube_eqn.items():
            s = iscale.get_scaling_factor(self.N_Nu_tube[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        for t, c in self.N_Nu_shell_eqn.items():
            s = iscale.get_scaling_factor(self.N_Nu_shell[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        for t, c in self.hconv_shell_total_eqn.items():
            s = iscale.get_scaling_factor(
                self.hconv_shell_total[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        if hasattr(self, "deltaP_shell_eqn"):
            for t, c in self.deltaP_shell_eqn.items():
                s = iscale.get_scaling_factor(
                    self.deltaP_shell[t], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s, overwrite=False)

        if hasattr(self, "deltaP_tube_eqn"):
            for t, c in self.deltaP_tube_eqn.items():
                s = iscale.get_scaling_factor(
                    self.deltaP_tube[t], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s, overwrite=False)

        if hasattr(self, "deltaP_tube_friction_eqn"):
            for t, c in self.deltaP_tube_friction_eqn.items():
                s = iscale.get_scaling_factor(
                    self.deltaP_tube_friction[t], default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np
from math import sqrt
from idaes.apps.matopt.materials import (
    Atom,
    Canvas,
    Design,
    LinearTiling,
    PlanarTiling,
    CubicTiling,
)
from idaes.apps.matopt.materials.geometry import (
    Shape,
    Cuboctahedron,
    Parallelepiped,
    Rhombohedron,
    RectPrism,
    Cube,
    Cylinder,
    CylindricalSector,
)
from idaes.apps.matopt.materials.lattices import (
    FCCLattice,
    CubicLattice,
    PerovskiteLattice,
    DiamondLattice,
    WurtziteLattice,
)
from idaes.apps.matopt.materials.transform_func import (
    ShiftFunc,
    ScaleFunc,
    RotateFunc,
    ReflectFunc,
)
from idaes.apps.matopt.opt import (
    Coef,
    LinearExpr,
    MatOptModel,
    SumNeighborSites,
    SumNeighborBonds,
    SumSites,
    SumBonds,
    SumSiteTypes,
    SumBondTypes,
    SumSitesAndTypes,
    SumBondsAndTypes,
    SumConfs,
    SumSitesAndConfs,
    LessThan,
    EqualTo,
    GreaterThan,
    FixedTo,
    Disallow,
    PiecewiseLinear,
    Implies,
    NegImplies,
    ImpliesSiteCombination,
    ImpliesNeighbors,
)
import pytest


@pytest.mark.unit
def test_construct_FCCLattice():
    IAD = sqrt(2) / 2
    lattice = FCCLattice.alignedWith100(IAD)
    lattice = FCCLattice.alignedWith111(IAD)
    lattice = FCCLattice(IAD)
    return lattice


@pytest.mark.unit
def test_construct_CubicLattice():
    IAD = 1.0
    lattice = CubicLattice(IAD)
    return lattice


@pytest.mark.unit
def test_construct_PerovskiteLattice():
    A, B, C = 1.0, 1.0, 1.0
    lattice = PerovskiteLattice(A, B, C)
    return lattice


@pytest.mark.unit
def test_construct_DiamondLattice():
    IAD = sqrt(3) / 4
    lattice = DiamondLattice.alignedWith(IAD, "100")
    lattice = DiamondLattice.alignedWith(IAD, "110")
    lattice = DiamondLattice.alignedWith(IAD, "111")
    lattice = DiamondLattice(IAD)
    return lattice


@pytest.mark.unit
def test_construct_WurtziteLattice():
    IAD = sqrt(3 / 8)
    lattice = WurtziteLattice.alignedWith(IAD, "0001")
    lattice = WurtziteLattice(IAD)
    return lattice


@pytest.mark.unit
def test_construct_Atom():
    atom_zero = Atom(39)
    atom_one = Atom(40)
    atom_two = Atom("Zr")
    assert atom_one == atom_two
    return atom_zero, atom_one


@pytest.mark.unit
def test_construct_Canvas():
    canvas = Canvas()
    return canvas


@pytest.mark.unit
def test_construct_Design():
    lattice = test_construct_FCCLattice()
    canvas = test_construct_Canvas()
    atom, _ = test_construct_Atom()
    canvas.addLocation(np.array([0, 0, 0], dtype=float))
    canvas.addShell(lattice.getNeighbors)
    canvas.setNeighborsFromFunc(lattice.getNeighbors)
    design = Design(canvas, atom)
    return design


@pytest.mark.unit
def test_construct_Shape():
    shape = Shape(np.array([0, 0, 0], dtype=float))
    return shape


@pytest.mark.unit
def test_construct_Cuboctahedron():
    shape = Cuboctahedron(1.0)
    return shape


@pytest.mark.unit
def test_construct_Parallelepiped():
    shape = Parallelepiped(
        np.array([1, 0, 0], dtype=float),
        np.array([0, 1, 0], dtype=float),
        np.array([0, 0, 1], dtype=float),
    )
    return shape


@pytest.mark.unit
def test_construct_Rhombohedron():
    shape = Rhombohedron(1.0, 0.5)
    return shape


@pytest.mark.unit
def test_construct_RectPrism():
    shape = RectPrism(1.0, 1.0, 1.0)
    return shape


@pytest.mark.unit
def test_construct_Cube():
    shape = Cube(1.0)
    return shape


@pytest.mark.unit
def test_construct_Cylinder():
    shape = Cylinder(np.zeros(3, dtype=float), 1.0, 1.0)
    return shape


@pytest.mark.unit
def test_construct_CylindricalSector():
    shape = CylindricalSector(
        np.zeros(3, dtype=float),
        1.0,
        1.0,
        np.array([1, 0, 0], dtype=float),
        np.array([0, 1, 0], dtype=float),
        np.array([0, 0, 1], dtype=float),
    )
    return shape


@pytest.mark.unit
def test_construct_LinearTiling():
    parallelepiped = test_construct_Parallelepiped()
    cylinder = test_construct_Cylinder()
    tiling = LinearTiling.fromParallelepiped(parallelepiped)
    tiling = LinearTiling.fromCylindricalShape(cylinder)
    return tiling


@pytest.mark.unit
def test_construct_PlanarTiling():
    shape = test_construct_Parallelepiped()
    tiling = PlanarTiling(shape)
    return tiling


@pytest.mark.unit
def test_construct_CubicTiling():
    shape = test_construct_Parallelepiped()
    tiling = CubicTiling(shape)
    return tiling


@pytest.mark.unit
def test_construct_ShiftFunc():
    transformation = ShiftFunc(np.array([1, 0, 0], dtype=float))
    return transformation


@pytest.mark.unit
def test_construct_ScaleFunc():
    transformation = ScaleFunc(2.0)
    return transformation


@pytest.mark.unit
def test_construct_RotateFunc():
    transformation = RotateFunc.fromXYZAngles(0.5, 0.5, 0.5)
    return transformation


@pytest.mark.unit
def test_construct_ReflectFunc():
    transformation = ReflectFunc.acrossX()
    return transformation


@pytest.mark.unit
def test_construct_Coef():
    coefficient = Coef([0.5, 0.5])
    return coefficient


@pytest.mark.unit
def test_construct_LinearExpr():
    expression = LinearExpr()
    return expression


@pytest.mark.unit
def test_construct_SumNeighborSites():
    m = test_construct_MatOptModel()
    expression = SumNeighborSites(m.Yi)
    return expression


@pytest.mark.unit
def test_construct_SumNeighborBonds():
    m = test_construct_MatOptModel()
    expression = SumNeighborBonds(m.Xij)
    return expression


@pytest.mark.unit
def test_construct_SumSites():
    m = test_construct_MatOptModel()
    expression = SumSites(m.Yi)
    return expression


@pytest.mark.unit
def test_construct_SumBonds():
    m = test_construct_MatOptModel()
    expression = SumBonds(m.Xij)
    return expression


@pytest.mark.unit
def test_construct_SumSiteTypes():
    m = test_construct_MatOptModel()
    expression = SumSiteTypes(m.Yik)
    return expression


@pytest.mark.unit
def test_construct_SumBondTypes():
    m = test_construct_MatOptModel()
    expression = SumBondTypes(m.Xijkl)
    return expression


@pytest.mark.unit
def test_construct_SumSitesAndTypes():
    m = test_construct_MatOptModel()
    expression = SumSitesAndTypes(m.Yik)
    return expression


@pytest.mark.unit
def test_construct_SumBondsAndTypes():
    m = test_construct_MatOptModel()
    expression = SumBondsAndTypes(m.Xijkl)
    return expression


@pytest.mark.unit
def test_construct_SumConfs():
    lattice = test_construct_FCCLattice()
    canvas = test_construct_Canvas()
    canvas.addLocation(np.array([0, 0, 0], dtype=float))
    canvas.addShell(lattice.getNeighbors)
    canvas.setNeighborsFromFunc(lattice.getNeighbors)
    confs = [[None] * len(canvas.NeighborhoodIndexes[0]) for _ in range(1)]
    a0, a1 = test_construct_Atom()
    atoms = [a0, a1]
    m = MatOptModel(canvas, atoms, confs)
    expression = SumConfs(m.Zic)
    return expression


@pytest.mark.unit
def test_construct_SumSitesAndConfs():
    lattice = test_construct_FCCLattice()
    canvas = test_construct_Canvas()
    canvas.addLocation(np.array([0, 0, 0], dtype=float))
    canvas.addShell(lattice.getNeighbors)
    canvas.setNeighborsFromFunc(lattice.getNeighbors)
    confs = [[None] * len(canvas.NeighborhoodIndexes[0]) for _ in range(1)]
    a0, a1 = test_construct_Atom()
    atoms = [a0, a1]
    m = MatOptModel(canvas, atoms, confs)
    expression = SumSitesAndConfs(m.Zic)
    return expression


@pytest.mark.unit
def test_construct_LessThan():
    rule = LessThan(2)
    return rule


@pytest.mark.unit
def test_construct_EqualTo():
    rule = EqualTo(2)
    return rule


@pytest.mark.unit
def test_construct_GreaterThan():
    rule = GreaterThan(2)
    return rule


@pytest.mark.unit
def test_construct_FixedTo():
    rule = FixedTo(2)
    return rule


@pytest.mark.unit
def test_construct_Disallow():
    design = test_construct_Design()
    rule = Disallow(design)
    return rule


@pytest.mark.unit
def test_construct_PiecewiseLinear():
    m = test_construct_MatOptModel()
    rule = PiecewiseLinear([x**2 for x in range(3)], [x for x in range(3)], m.Ci)
    return rule


@pytest.mark.unit
def test_construct_Implies():
    m = test_construct_MatOptModel()
    r = test_construct_LessThan()
    rule = Implies((m.Yi, r))
    return rule


@pytest.mark.unit
def test_construct_NegImplies():
    m = test_construct_MatOptModel()
    r = test_construct_LessThan()
    rule = NegImplies((m.Yi, r))
    return rule


@pytest.mark.unit
def test_construct_ImpliesSiteCombination():
    canvas = test_construct_Canvas()
    m = test_construct_MatOptModel()
    r = test_construct_GreaterThan()
    rule = ImpliesSiteCombination(canvas, (m.Yi, r), (m.Yik, r))
    return rule


@pytest.mark.unit
def test_construct_ImpliesNeighbors():
    m = test_construct_MatOptModel()
    r = test_construct_FixedTo()
    rule = ImpliesNeighbors((m.Yi, r))
    return rule


@pytest.mark.unit
def test_construct_MatOptModel():
    canvas = test_construct_Canvas()
    a0, a1 = test_construct_Atom()
    atoms = [a0, a1]
    model = MatOptModel(canvas, atoms)
    return model



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base class for EoS modules.

Raises NotImplementedErrors for all expected methods in case developer misses
some. EoS developers should overload all these methods.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import units as pyunits
from idaes.core.util.constants import Constants as const
from idaes.models.properties.modular_properties.base.utility import (
    get_method,
    get_component_object as cobj,
)
from idaes.core.util.exceptions import ConfigurationError


class EoSBase:
    """Base class for modular equation-of-state classes."""

    @staticmethod
    def gas_constant(b):
        # Utility method to convert gas constant to base units
        base_units = b.params.get_metadata().default_units

        return pyunits.convert(const.gas_constant, to_units=base_units.GAS_CONSTANT)

    @staticmethod
    def common(b, pobj):
        raise NotImplementedError(_msg(b, "common"))

    @staticmethod
    def calculate_scaling_factors(b, pobj):
        raise NotImplementedError(_msg(b, "calculate_scaling_factors"))

    @staticmethod
    def build_parameters(b):
        raise NotImplementedError(_msg(b, "build_parameters"))

    @staticmethod
    def build_critical_properties(b, ref_phase):
        """
        This method is used to define the constraints used to calculate mixture
        critical properties.
        """
        raise NotImplementedError(_msg(b, "build_critical_properties"))

    @staticmethod
    def list_critical_property_constraint_names():
        """
        If critical properties are supported, this method is required during
        initialization to identify the constraints that are associated
        with calculating the critical properties.

        Returns:
            list of constraint names
        """
        raise NotImplementedError(
            "Equation of State module has not implemented a method for "
            "list_critical_property_constraint_names. Please contact the "
            "EoS developer or use a different module."
        )

    @staticmethod
    def get_vol_mol_pure(b, phase, comp, temperature):
        try:
            vol_mol = get_method(b, "vol_mol_" + phase + "_comp", comp)(
                b, cobj(b, comp), temperature
            )
        except (AttributeError, ConfigurationError):
            # vol_mol not defined, try for dens_mol instead
            try:
                vol_mol = 1 / get_method(b, "dens_mol_" + phase + "_comp", comp)(
                    b, cobj(b, comp), temperature
                )
            except (AttributeError, ConfigurationError):
                # Does not have either vol_mol or dens_mol
                suffix = "_" + phase + "_comp"
                raise ConfigurationError(
                    f"{b.name} does not have a method defined to use "
                    f"when calculating molar volume and density for "
                    f"component {comp} in phase {phase}. Each component "
                    f"must define a method for either vol_mol{suffix} or "
                    f"dens_mol{suffix}."
                )
        return vol_mol

    @staticmethod
    def act_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "act_phase_comp"))

    @staticmethod
    def act_phase_comp_true(b, p, j):
        raise NotImplementedError(_msg(b, "act_phase_comp_true"))

    @staticmethod
    def act_phase_comp_appr(b, p, j):
        raise NotImplementedError(_msg(b, "act_phase_comp_appr"))

    @staticmethod
    def act_coeff_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "act_coeff_phase_comp"))

    @staticmethod
    def act_coeff_phase_comp_true(b, p, j):
        raise NotImplementedError(_msg(b, "act_coeff_phase_comp_true"))

    @staticmethod
    def act_coeff_phase_comp_appr(b, p, j):
        raise NotImplementedError(_msg(b, "act_coeff_phase_comp_appr"))

    @staticmethod
    def cp_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "cp_mol_phase"))

    @staticmethod
    def cp_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "cp_mol_phase_comp"))

    @staticmethod
    def cv_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "cv_mol_phase"))

    @staticmethod
    def cv_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "cv_mol_phase_comp"))

    @staticmethod
    def heat_capacity_ratio_phase(b, p):
        return b.cp_mol_phase[p] / b.cv_mol_phase[p]

    @staticmethod
    def cv_mol_ig_comp_pure(b, j):
        # Method for calculating pure component ideal gas cv from cp
        # For ideal gases, cv = cp - R
        units = b.params.get_metadata().derived_units
        R = pyunits.convert(const.gas_constant, to_units=units.HEAT_CAPACITY_MOLE)
        return get_method(b, "cp_mol_ig_comp", j)(b, cobj(b, j), b.temperature) - R

    @staticmethod
    def cv_mol_ls_comp_pure(b, p, j):
        # Method for calculating pure component liquid and solid cv from cp
        # For ideal (incompressible) liquids and solids, cv = cp
        pobj = b.params.get_phase(p)
        if pobj.is_liquid_phase():
            return get_method(b, "cp_mol_liq_comp", j)(b, cobj(b, j), b.temperature)
        elif pobj.is_solid_phase():
            return get_method(b, "cp_mol_sol_comp", j)(b, cobj(b, j), b.temperature)

    @staticmethod
    def dens_mass_phase(b, p):
        raise NotImplementedError(_msg(b, "dens_mass_phase"))

    @staticmethod
    def dens_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "dens_mol_phase"))

    @staticmethod
    def energy_internal_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "energy_internal_mol_phase"))

    @staticmethod
    def energy_internal_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "energy_internal_mol_phase_comp"))

    @staticmethod
    def energy_internal_mol_ig_comp_pure(b, j):
        # Method for calculating pure component U from H for ideal gases
        units = b.params.get_metadata().derived_units
        R = pyunits.convert(const.gas_constant, to_units=units.HEAT_CAPACITY_MOLE)

        if cobj(b, j).parent_block().config.include_enthalpy_of_formation:
            # First, need to determine correction between U_form and H_form
            # U_form = H_form - delta_n*R*T
            ele_comp = cobj(b, j).config.elemental_composition
            if ele_comp is None:
                raise ConfigurationError(
                    "{} calculation of internal energy requires elemental "
                    "composition of all species. Please set this using the "
                    "elemental_composition argument in the component "
                    "declaration ({}).".format(b.name, j)
                )

            delta_n = 0
            for e, s in ele_comp.items():
                # Check for any element which is vapor at standard state
                if e in ["He", "Ne", "Ar", "Kr", "Xe", "Ra"]:
                    delta_n += -s
                elif e in ["F", "Cl", "H", "N", "O"]:
                    delta_n += -s / 2  # These are diatomic at standard state

            delta_n += 1  # One mole of gaseous compound is formed
            dU_form = delta_n * R * b.params.temperature_ref
        else:
            dU_form = 0  # No heat of formation to correct

        # For ideal gases, U = H - R(T-T_ref) + dU_form
        return (
            get_method(b, "enth_mol_ig_comp", j)(b, cobj(b, j), b.temperature)
            - R * (b.temperature - b.params.temperature_ref)
            + dU_form
        )

    @staticmethod
    def energy_internal_mol_ls_comp_pure(b, p, j):
        pobj = b.params.get_phase(p)
        if pobj.is_liquid_phase():
            mthd = get_method(b, "enth_mol_liq_comp", j)
        elif pobj.is_solid_phase():
            mthd = get_method(b, "enth_mol_sol_comp", j)

        # Method for calculating pure component U from H for liquids & solids
        units = b.params.get_metadata().derived_units
        R = pyunits.convert(const.gas_constant, to_units=units.HEAT_CAPACITY_MOLE)

        if cobj(b, j).parent_block().config.include_enthalpy_of_formation:
            # First, need to determine correction between U_form and H_form
            # U_form = H_form - delta_n*R*T
            ele_comp = cobj(b, j).config.elemental_composition
            if ele_comp is None:
                raise ConfigurationError(
                    "{} calculation of internal energy requires elemental "
                    "composition of all species. Please set this using the "
                    "elemental_composition argument in the component "
                    "declaration ({}).".format(b.name, j)
                )

            delta_n = 0
            for e, s in ele_comp.items():
                # Check for any element which is vapor at standard state
                if e in ["He", "Ne", "Ar", "Kr", "Xe", "Ra"]:
                    delta_n += -s
                elif e in ["F", "Cl", "H", "N", "O"]:
                    delta_n += -s / 2  # These are diatomic at standard state
            dU_form = delta_n * R * b.params.temperature_ref

            # For ideal (incompressible) liquids and solids, U = H + dU_form
            return mthd(b, cobj(b, j), b.temperature) + dU_form
        else:
            # If not including heat of formation, U = H
            return mthd(b, cobj(b, j), b.temperature)

    @staticmethod
    def enth_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "enth_mol_phase"))

    @staticmethod
    def enth_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "enth_mol_phase_comp"))

    @staticmethod
    def entr_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "entr_mol_phase"))

    @staticmethod
    def entr_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "entr_mol_phase_comp"))

    @staticmethod
    def fug_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "fug_phase_comp"))

    @staticmethod
    def fug_phase_comp_eq(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_eq"))

    @staticmethod
    def log_fug_phase_comp_eq(b, p, j, pp):
        raise NotImplementedError(_msg(b, "log_fug_phase_comp_eq"))

    @staticmethod
    def fug_coeff_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "fug_coeff_phase_comp"))

    @staticmethod
    def fug_coeff_phase_comp_eq(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_coeff_phase_comp_eq"))

    @staticmethod
    def fug_phase_comp_Tbub(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Tbub"))

    @staticmethod
    def fug_phase_comp_Tdew(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Tdew"))

    @staticmethod
    def fug_phase_comp_Pbub(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Pbub"))

    @staticmethod
    def fug_phase_comp_Pdew(b, p, j, pp):
        raise NotImplementedError(_msg(b, "fug_phase_comp_Pdew"))

    @staticmethod
    def gibbs_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "gibbs_mol_phase"))

    @staticmethod
    def gibbs_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "gibbs_mol_phase_comp"))

    @staticmethod
    def isentropic_speed_sound_phase(b, p):
        raise NotImplementedError(_msg(b, "isentropic_speed_sound_phase"))

    @staticmethod
    def isothermal_speed_sound_phase(b, p):
        raise NotImplementedError(_msg(b, "isothermal_speed_sound_phase"))

    def pressure_osm_phase(b, p):
        raise NotImplementedError(_msg(b, "pressure_osm_phase"))

    @staticmethod
    def vol_mol_phase(b, p):
        raise NotImplementedError(_msg(b, "vol_mol_phase"))

    @staticmethod
    def vol_mol_phase_comp(b, p, j):
        raise NotImplementedError(_msg(b, "vol_mol_phase_comp"))


def _msg(b, attr):
    return (
        "{} Equation of State module has not implemented a method for {}. "
        "Please contact the EoS developer or use a different module.".format(
            b.name, attr
        )
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pyomo.environ as pyo
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError
import matplotlib.pyplot as plt
import logging

_logger = logging.getLogger(__name__)

no_init_func_message = "initialization_func argument was not provided. Returning the multiperiod model without initialization."


class MultiPeriodModel(pyo.ConcreteModel):
    """
    The `MultiPeriodModel` class helps transfer existing steady-state
    process models to multiperiod versions that contain dynamic time coupling.

    Arguments:
        n_time_points: number of points to use in time horizon
        process_model_func: function that returns a multiperiod capable pyomo model
        linking_variable_func: function that returns a tuple of variable
                               pairs to link between time steps
        periodic_variable_func: a function that returns a tuple of variable
                                pairs to link between last and first time steps
        use_stochastic_build: Uses `build_stochastic_multi_period` method if set to True
        set_days: list containing the set of representative days
        set_years: list containing the set of years
        set_scenarios: list containing the set of scenarios
        initialization_func: function that fixes the degrees of freedom and initializes an
                             instance of the flowsheet.
        unfix_dof_func: function that unfixes a few degrees of freedom for optimization
        flowsheet_options: dictionary containing the arguments needed for `process_model_func`
        initialization_options: dictionary containing the arguments needed for `initialization_func`
        unfix_dof_options: dictionary containing the arguments needed for `unfix_dof_func`
        solver: pyomo solver object
        outlvl: logging level

    Returns:
        (stochastic) multi-period optimization model
    """

    def __init__(
        self,
        n_time_points,
        process_model_func,
        linking_variable_func,
        periodic_variable_func=None,
        use_stochastic_build=False,
        set_days=None,
        set_years=None,
        set_scenarios=None,
        initialization_func=None,
        unfix_dof_func=None,
        flowsheet_options=None,
        initialization_options=None,
        unfix_dof_options=None,
        solver=None,
        outlvl=logging.WARNING,
    ):  # , state_variable_func=None):
        super().__init__()

        if flowsheet_options is None:
            flowsheet_options = {}
        if initialization_options is None:
            initialization_options = {}
        if unfix_dof_options is None:
            unfix_dof_options = {}
        self.n_time_points = n_time_points

        # user provided functions
        self.create_process_model = process_model_func
        self.get_linking_variable_pairs = linking_variable_func
        self.get_periodic_variable_pairs = periodic_variable_func
        self.initialization_func = initialization_func
        self.unfix_dof_func = unfix_dof_func
        # self.get_state_variable_pairs = state_variable_func

        # populated on 'build_multi_period_model'
        self._first_active_time = None

        # Create sets
        if use_stochastic_build:
            self.set_time = pyo.RangeSet(n_time_points)

            if set_days is not None:
                self.set_days = pyo.Set(initialize=set_days)
                self._multiple_days = True

            else:
                self._multiple_days = False

            if set_years is not None:
                self.set_years = pyo.Set(initialize=set_years)
                self._multiyear = True

            else:
                self._multiyear = False

            if set_scenarios is not None:
                self.set_scenarios = pyo.Set(initialize=set_scenarios)
                self._stochastic_model = True

            else:
                self._stochastic_model = False

            if solver is None:
                solver = get_solver()

            _logger = logging.getLogger(__name__)
            _logger.setLevel(outlvl)

            # Build the stochastic multiperiod optimization model
            self.build_stochastic_multi_period(
                flowsheet_options,
                initialization_options,
                unfix_dof_options,
                solver,
            )

        # optional initialization features
        # self.initialization_points = None   #library of possible initial points
        # self.initialize_func = None         #function to perform the initialize

    def build_multi_period_model(
        self,
        model_data_kwargs=None,
        flowsheet_options=None,
        initialization_options=None,
        unfix_dof_options=None,
        solver=None,
    ):
        """
        Build a multi-period capable model using user-provided functions

        Arguments:
            model_data_kwargs: a dict of dicts with {time:{"key",value}}
                               where `time` is the time in the horizon. each
                               `time` dictionary is passed to the
                               `create_process_model` function
            flowsheet_options: dict containing the arguments needed to build an instance of flowsheet
            initialization_options: dict containing the arguments needed for `initialization_func`
            unfix_dof_options: dict containing the arguments needed for `unfix_dof_func`
            solver: pyomo solver object
        """
        if flowsheet_options is None:
            flowsheet_options = {}
        if initialization_options is None:
            initialization_options = {}
        if unfix_dof_options is None:
            unfix_dof_options = {}
        if solver is None:
            solver = get_solver()

        # Challenge: If model_data_kwargs is provided, then we cannot use
        # the initialize_multi_period_model method because that will overwrite
        # the parameter values. In that case, we may have to call initialization
        # function for each instance of the flowsheet. Not sure if this method
        # will be required or not in general, so this is what I'm going to do:
        # If the argument is not provided, then use clone and initialize. If it
        # is provided, then return the multiperiod model without initialization.

        m = self
        m.TIME = pyo.Set(initialize=range(self.n_time_points))
        m.blocks = pyo.Block(m.TIME)

        if model_data_kwargs is None:
            blk = self._construct_flowsheet_instance(
                flowsheet_options=flowsheet_options,
                initialization_options=initialization_options,
                unfix_dof_options=unfix_dof_options,
                solver=solver,
            )
            for t in m.TIME:
                _logger.info(f"Constructing flowsheet model for time index {t}")
                m.blocks[t].process = blk.clone()

        else:
            if len(model_data_kwargs) != self.n_time_points:
                _logger.error(
                    f"len(model_data_kwargs) != n_time_points.\n "
                    f"len(model_data_kwargs) = {len(model_data_kwargs)}\n"
                    f"len(n_time_points) = {self.n_time_points}\n"
                    f"Check input data for model_data_kwargs argument."
                )

            _logger.warning(
                f"model_data_kwargs argument is provided, so the flowsheet "
                f"options are different for different time instances. In this case, "
                f"the multiperiod model is returned without initialization."
            )
            for t in m.TIME:
                _logger.info(f"Constructing flowsheet model for time index {t}")
                m.blocks[t].process = self.create_process_model(**model_data_kwargs[t])

        # link blocks together. loop over every time index except the last one
        for t in m.TIME.data()[: self.n_time_points - 1]:
            link_variable_pairs = self.get_linking_variable_pairs(
                m.blocks[t].process, m.blocks[t + 1].process
            )
            self._create_linking_constraints(m.blocks[t].process, link_variable_pairs)

        if self.get_periodic_variable_pairs is not None:
            N = len(m.blocks)
            periodic_variable_pairs = self.get_periodic_variable_pairs(
                m.blocks[N - 1].process, m.blocks[0].process
            )
            self._create_periodic_constraints(
                m.blocks[N - 1].process, periodic_variable_pairs
            )

        self._first_active_time = m.TIME.first()
        return m

    def advance_time(self, **model_data_kwargs):
        """
        Advance the current model instance to the next time period

        Arguments:
            model_data_kwargs: keyword arguments passed to user provided
                               `create_process_model` function
        """
        m = self
        previous_time = self._first_active_time
        current_time = m.TIME.next(previous_time)

        # deactivate previous time
        m.blocks[previous_time].process.deactivate()

        # track the first time in the problem horizon
        self._first_active_time = current_time

        # populate new time for the end of the horizon
        last_time = m.TIME.last()
        new_time = last_time + 1
        m.TIME.add(new_time)
        m.blocks[new_time].process = self.create_process_model(**model_data_kwargs)

        # sequential time coupling
        link_variable_pairs = self.get_linking_variable_pairs(
            m.blocks[last_time].process, m.blocks[new_time].process
        )
        self._create_linking_constraints(
            m.blocks[last_time].process, link_variable_pairs
        )

        # periodic time coupling
        if self.get_periodic_variable_pairs is not None:
            periodic_variable_pairs = self.get_periodic_variable_pairs(
                m.blocks[new_time].process, m.blocks[current_time].process
            )
            self._create_periodic_constraints(
                m.blocks[new_time].process, periodic_variable_pairs
            )
            # deactivate old periodic constraint
            m.blocks[last_time].process.periodic_constraints.deactivate()

        # TODO: discuss where state goes.
        # sometimes the user might want to fix values based on a 'real' process
        # also TODO: inspect argument and use fix() if possible
        # if self.get_state_variable_pairs is not None:
        #     state_variable_pairs = self.get_state_variable_pairs(
        #                       m.blocks[previous_time].process,
        #                       m.blocks[current_time].process)
        #     self._fix_initial_states(
        #                       m.blocks[current_time].process,
        #                       state_variable_pairs)

    @property
    def pyomo_model(self):
        """
        Retrieve the underlying pyomo model
        """
        return self

    @property
    def current_time(self):
        """
        Retrieve the current multiperiod model time
        """
        return self._first_active_time

    def get_active_process_blocks(self):
        """
        Retrieve the active time blocks of the pyomo model
        """
        return [b.process for b in self.blocks.values() if b.process.active]

    def _create_linking_constraints(self, b1, variable_pairs):
        """
        Create linking constraint on `b1` using `variable_pairs`
        """
        b1.link_constraints = pyo.Constraint(range(len(variable_pairs)))
        for i, pair in enumerate(variable_pairs):
            b1.link_constraints[i] = pair[0] == pair[1]

    def _create_periodic_constraints(self, b1, variable_pairs):
        """
        Create periodic linking constraint on `b1` using `variable_pairs`
        """
        b1.periodic_constraints = pyo.Constraint(range(len(variable_pairs)))
        for i, pair in enumerate(variable_pairs):
            b1.periodic_constraints[i] = pair[0] == pair[1]

    def _construct_flowsheet_instance(
        self,
        flowsheet_options,
        initialization_options,
        unfix_dof_options,
        solver,
    ):
        # Create an instance of the flowsheet for cloning
        blk = pyo.ConcreteModel()
        self.create_process_model(blk, **flowsheet_options)

        if self.initialization_func is None:
            _logger.warning(no_init_func_message)

        else:
            self.initialization_func(blk, **initialization_options)
            result = solver.solve(blk)

            if not pyo.check_optimal_termination(result):
                raise InitializationError(
                    "Flowsheet did not converge after fixing the degrees of freedom. "
                    "To create the multi-period model without initialization, do not provide "
                    "initialization_func argument."
                )

        if self.unfix_dof_func is None:
            _logger.warning(
                "unfix_dof_func argument is not provided. "
                "Returning the model without unfixing degrees of freedom."
            )

        else:
            self.unfix_dof_func(blk, **unfix_dof_options)

        return blk

    def build_stochastic_multi_period(
        self,
        flowsheet_options,
        initialization_options,
        unfix_dof_options,
        solver,
    ):
        """
        This function constructs the stochastic multiperiod optimization problem
        """
        # Create set of periods:
        multiple_days = self._multiple_days
        multiyear = self._multiyear

        # Construct the set of time periods
        if multiyear and multiple_days:
            set_period = [
                (t, d, y)
                for y in self.set_years
                for d in self.set_days
                for t in self.set_time
            ]

        elif multiyear and not multiple_days:
            set_period = [(t, y) for y in self.set_years for t in self.set_time]

        elif not multiyear and multiple_days:
            set_period = [(t, d) for d in self.set_days for t in self.set_time]

        else:
            set_period = [t for t in self.set_time]

        self.set_period = pyo.Set(initialize=set_period)

        def _build_scenario_model(m, fs_blk):
            """
            Construct a multiperiod model for one scenario

            Arguments:
                m: pyomo concrete model for the MultiPeriod model
                fs_blk: flowsheet block to be cloned to each time index
            """

            m.period = pyo.Block(self.set_period)

            for i in m.period:
                _logger.info(f"Constructing the flowsheet model for index {i}")

                m.period[i].transfer_attributes_from(fs_blk.clone())

            # link blocks together. loop over every time index except the last one
            if self.get_linking_variable_pairs is None:
                _logger.warning(
                    "linking_variable_func is not provided, so variables across"
                    " time periods are not linked."
                )
                return

            # TODO: Instead of accessing the constraints as
            # m.link_constraints[t, d, y].link_constraints[1], add a Reference for easy access.
            if multiyear and multiple_days:
                m.link_constraints = pyo.Block(
                    self.set_time.data()[:-1], self.set_days, self.set_years
                )

                for y in self.set_years:
                    for d in self.set_days:
                        for t in self.set_time.data()[:-1]:
                            link_variable_pairs = self.get_linking_variable_pairs(
                                m.period[t, d, y], m.period[t + 1, d, y]
                            )
                            self._create_linking_constraints(
                                m.link_constraints[t, d, y], link_variable_pairs
                            )

            elif multiyear and not multiple_days:
                m.link_constraints = pyo.Block(
                    self.set_time.data()[:-1], self.set_years
                )

                for y in self.set_years:
                    for t in self.set_time.data()[:-1]:
                        link_variable_pairs = self.get_linking_variable_pairs(
                            m.period[t, y], m.period[t + 1, y]
                        )
                        self._create_linking_constraints(
                            m.link_constraints[t, y], link_variable_pairs
                        )

            elif not multiyear and multiple_days:
                m.link_constraints = pyo.Block(self.set_time.data()[:-1], self.set_days)

                for d in self.set_days:
                    for t in self.set_time.data()[:-1]:
                        link_variable_pairs = self.get_linking_variable_pairs(
                            m.period[t, d], m.period[t + 1, d]
                        )
                        self._create_linking_constraints(
                            m.link_constraints[t, d], link_variable_pairs
                        )

            else:
                m.link_constraints = pyo.Block(self.set_time.data()[:-1])

                for t in self.set_time.data()[:-1]:
                    link_variable_pairs = self.get_linking_variable_pairs(
                        m.period[t], m.period[t + 1]
                    )
                    self._create_linking_constraints(
                        m.link_constraints[t], link_variable_pairs
                    )

            # Check if a method for periodic constraints is given
            if self.get_periodic_variable_pairs is not None:
                _logger.warning(
                    "A method is provided for get_periodic_variable_pairs. "
                    "build_stochastic_multi_period method does not support periodic "
                    "constraints, so the user needs to add them manually."
                )

        # Create an instance of the flowsheet
        blk = self._construct_flowsheet_instance(
            flowsheet_options=flowsheet_options,
            initialization_options=initialization_options,
            unfix_dof_options=unfix_dof_options,
            solver=solver,
        )

        # Begin the formulation of the multiperiod optimization problem
        if self._stochastic_model:
            self.scenario = pyo.Block(self.set_scenarios)
            sce_blk = pyo.ConcreteModel()
            _build_scenario_model(sce_blk, blk)

            for i in self.scenario:
                _logger.info(f"Constructing the model for scenario {i}")
                self.scenario[i].transfer_attributes_from(sce_blk.clone())

        else:
            _build_scenario_model(self, blk)

    @staticmethod
    def plot_lmp_signal(
        lmp, time=None, draw_style="steps", x_range=None, y_range=None, grid=None
    ):
        """
        This function plots LMP signals as a function of time.

        Args:
            lmp: list or dict of LMP signals (length of dictionary must be <= 6).
            time: list or dict of time. Taken to be 1:len(lmp) if unspecified.
            draw_style: Plot style. Should be either "steps" or "default".
            x_range: tuple or dict of tuples with x-axis range.
            y_range: tuple or dict of tuples with y-axis range.
            grid: Grid shape of the plot.

        Returns:
            Plot of LMP signals
        """

        if type(lmp) is list:
            grid = (1, 1)

            if time is None:
                plt_time = {1: [i for i in range(1, len(lmp) + 1)]}

            else:
                plt_time = {1: time}

            plt_lmp = {1: lmp}
            plt_title = {1: ""}
            color = {1: "tab:red"}
            plt_x_range = {1: x_range}
            plt_y_range = {1: y_range}

        elif type(lmp) is dict:
            if len(lmp) > 6:
                raise Exception(
                    "Number of LMP signals provided exceeds six: the maximum "
                    "number of subplots the function supports."
                )

            grid_shape = {
                1: (1, 1),
                2: (1, 2),
                3: (2, 2),
                4: (2, 2),
                5: (2, 3),
                6: (2, 3),
            }
            if grid is None:
                grid = grid_shape[len(lmp)]

            color = {
                1: "tab:red",
                2: "tab:red",
                3: "tab:red",
                4: "tab:red",
                5: "tab:red",
                6: "tab:red",
            }

            plt_time = {}
            plt_lmp = {}
            plt_title = {}
            plt_x_range = {}
            plt_y_range = {}

            counter = 1
            for i in lmp:
                plt_lmp[counter] = lmp[i]

                if time is None:
                    plt_time[counter] = [j for j in range(1, len(lmp[i]) + 1)]

                else:
                    plt_time[counter] = time[i]

                plt_title[counter] = str(i)
                plt_x_range[counter] = (
                    None if x_range is None or i not in x_range else x_range[i]
                )
                plt_y_range[counter] = (
                    None if y_range is None or i not in y_range else y_range[i]
                )
                counter = counter + 1

        fig = plt.figure()

        for i in range(1, len(plt_lmp) + 1):
            ax = fig.add_subplot(grid[0], grid[1], i)
            ax.plot(plt_time[i], plt_lmp[i], color=color[i], drawstyle=draw_style)
            ax.set_xlabel("time (hr)")
            ax.set_ylabel("LMP ($/MWh)")
            ax.set_title(plt_title[i])

            if plt_x_range[i] is not None:
                ax.set_xlim(plt_x_range[i][0], plt_x_range[i][1])

            if plt_y_range[i] is not None:
                ax.set_ylim(plt_y_range[i][0], plt_y_range[i][1])

        fig.tight_layout()
        plt.show()

    @staticmethod
    def plot_lmp_and_schedule(
        lmp,
        schedule,
        time=None,
        y_label=None,
        x_range=None,
        lmp_range=None,
        y_range=None,
        x_label="time (hr)",
        lmp_label="LMP ($/MWh)",
        color=None,
        draw_style="steps",
        grid=None,
    ):
        """
        The function plots optimal operation schedule as a function of time.

        Args:
            lmp: list of LMPs.
            schedule: dict of operating schedules {"variable": [profile], ...}.
            time: list of time instances. Taken to be 1:len(lmp) if unspecified.
            y_label: dict of y labels for the schedule. Taken to be keys of schedule if unspecified.
            x_range: tuple containing the range of x-axis.
            lmp_range: tuple containing the range of lmp-axis.
            y_range: dict of tuples containing the range of y-axis.
            x_label: x-axis label. "time (hr)" is the default
            lmp_label: lmp-axis label. "LMP ($/MWh)" is the default.
            color: dict of colors for the plots.
            draw_style: plot style. Must be either "steps" or "default".
            grid: grid shape of the plost.

        Returns:
            None
        """
        if len(schedule) > 4:
            raise Exception(
                "Number of elements in schedule exceeds four: "
                "the maximum number of subplots the function supports."
            )

        key_list = {index + 1: value for index, value in enumerate(schedule)}

        if time is None:
            time = [i for i in range(1, len(lmp) + 1)]

        if grid is None:
            grid_shape = {1: (1, 1), 2: (1, 2), 3: (2, 2), 4: (2, 2)}
            grid = grid_shape[len(schedule)]

        lmp_color = "tab:red"
        if color is None:
            plt_color = {1: "tab:blue", 2: "magenta", 3: "tab:green", 4: "tab:cyan"}
        else:
            plt_color = {index + 1: color[value] for index, value in enumerate(color)}

        fig = plt.figure()

        for i in range(1, len(schedule) + 1):
            ax = fig.add_subplot(grid[0], grid[1], i)
            ax.set_xlabel(x_label)
            ax.set_ylabel(lmp_label, color=lmp_color)
            ax.plot(time, lmp, color=lmp_color, drawstyle="steps")
            ax.tick_params(axis="y", labelcolor=lmp_color)

            if x_range is not None:
                ax.set_xlim(x_range[0], x_range[1])

            if lmp_range is not None:
                ax.set_ylim(lmp_range[0], lmp_range[1])

            ax1 = ax.twinx()
            ax1.plot(
                time, schedule[key_list[i]], color=plt_color[i], drawstyle=draw_style
            )
            ax1.tick_params(axis="y", labelcolor=plt_color[i])

            if y_label is not None and key_list[i] in y_label:
                ax1.set_ylabel(y_label[key_list[i]], color=plt_color[i])

            else:
                ax1.set_ylabel(str(key_list[i]), color=plt_color[i])

            if y_range is not None and key_list[i] in y_range:
                ax1.set_ylim(y_range[key_list[i]][0], y_range[key_list[i]][1])

        fig.tight_layout()
        plt.show()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Watertank model test
The water tank has only one inlet and one outlet

main assumptions:
1.- Heat loss is a variable given by the user (zero heat loss can be
    specified if adiabatic)
2.- Calculate pressure change due to gravity based on water level
    and contraction to downcomer
3.- Water level is either fixed for steady-state model or calculated for
    dynamic model
4.- Assume enthalpy_in == enthalpy_out + heat loss
5.- Subcooled water from economizer and saturated water from waterwall
    are mixed before entering the tank

Created on Nov 04 2020 by J. Ma, M. Zamarripa, D. Caballero
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.unit_models.watertank import WaterTank

from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.fixture(scope="module")
def build_watertank_simple():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterTank(
        property_package=m.fs.prop_water,
        has_holdup=False,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    # fix inputs for simple tank
    m.fs.unit.tank_cross_sect_area.fix(1.14)  # tank cross sectional area
    m.fs.unit.tank_level[:].fix(0.6)  # tank level
    m.fs.unit.heat_duty[:].fix(0.0)  # assume no heat loss
    return m


@pytest.fixture(scope="module")
def build_watertank_rect():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterTank(
        tank_type="rectangular_tank",
        property_package=m.fs.prop_water,
        has_holdup=False,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    # fix inputs for horizontal cylindrical tank
    m.fs.unit.tank_width.fix(0.5)  # tank width
    m.fs.unit.tank_length.fix(1.2)  # tank length
    m.fs.unit.tank_level[:].fix(0.6)  # tank level
    m.fs.unit.heat_duty[:].fix(0.0)  # assume no heat loss
    return m


@pytest.fixture(scope="module")
def build_watertank_vert_cylin():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterTank(
        tank_type="vertical_cylindrical_tank",
        property_package=m.fs.prop_water,
        has_holdup=False,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    # fix inputs for horizontal cylindrical tank
    m.fs.unit.tank_diameter.fix(1.2)  # tank diameter
    m.fs.unit.tank_level[:].fix(0.6)  # tank level
    m.fs.unit.heat_duty[:].fix(0.0)  # assume no heat loss
    return m


@pytest.fixture(scope="module")
def build_watertank_hori_cylin():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterTank(
        tank_type="horizontal_cylindrical_tank",
        property_package=m.fs.prop_water,
        has_holdup=False,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    # fix inputs for horizontal cylindrical tank
    m.fs.unit.tank_diameter.fix(1.2)  # tank diameter
    m.fs.unit.tank_length.fix(15.3256)  # tank length
    m.fs.unit.tank_level[:].fix(0.6)  # tank level
    m.fs.unit.heat_duty[:].fix(0.0)  # assume no heat loss
    return m


@pytest.fixture(scope="module")
def tank_models(
    build_watertank_simple,
    build_watertank_rect,
    build_watertank_vert_cylin,
    build_watertank_hori_cylin,
):
    return [
        build_watertank_simple,
        build_watertank_rect,
        build_watertank_vert_cylin,
        build_watertank_hori_cylin,
    ]


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(tank_models):
    """Make a turbine model and make sure it doesn't throw exception"""
    for i in tank_models:
        # build the simple tank
        m = i
        assert degrees_of_freedom(m) == 3
        # Check unit config arguments
        assert len(m.fs.unit.config) == 10
        assert m.fs.unit.config.has_heat_transfer
        assert m.fs.unit.config.has_pressure_change
        assert m.fs.unit.config.property_package is m.fs.prop_water


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_watertank(tank_models):
    state_args_feedwater = {
        "flow_mol": 83193 / 3600 * 1000,  # mol/s
        "pressure": 12024201.99,  # Pa
        "enth_mol": 22723.907,
    }  # J/mol
    for i in tank_models:
        initialization_tester(i, dof=3, state_args=state_args_feedwater)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_run_watertank(tank_models):
    # volume_values = [simple tank, rectangular, vertical cylindrical,
    #                  horizontal cylindrical]
    volume_values = [0.684, 0.36, 0.679, 8.666]

    optarg = {"tol": 1e-7, "linear_solver": "ma27", "max_iter": 40}

    solver.options = optarg

    for i in tank_models:
        m = i

        # fix inlets
        m.fs.unit.inlet.flow_mol[:].fix()
        m.fs.unit.inlet.pressure[:].fix()
        m.fs.unit.inlet.enth_mol[:].fix()

        # solve model
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert pyo.check_optimal_termination(results)
        assert degrees_of_freedom(m) == 0
        assert pytest.approx(0.6, abs=1e-3) == pyo.value(m.fs.unit.tank_level[0])
        # mass balance
        assert pytest.approx(0, abs=1e-3) == pyo.value(
            m.fs.unit.inlet.flow_mol[0] - m.fs.unit.outlet.flow_mol[0]
        )
        # energy balance
        assert pytest.approx(0, abs=1e-3) == pyo.value(
            m.fs.unit.inlet.flow_mol[0] * m.fs.unit.inlet.enth_mol[0]
            - m.fs.unit.outlet.flow_mol[0] * m.fs.unit.outlet.enth_mol[0]
        )
        # pressure drop
        assert pytest.approx(4410.081, rel=1e-3) == pyo.value(m.fs.unit.deltaP[0])

        # volume
        assert pytest.approx(
            volume_values[tank_models.index(i)], abs=1e-3
        ) == pyo.value(m.fs.unit.volume[0])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Gibbs reactor Scaler.

Author: Andrew Lee
"""
import os
import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ComponentMap,
    ConcreteModel,
    Constraint,
    Suffix,
    TransformationFactory,
    units,
    value,
    Var,
)

from idaes.core import FlowsheetBlock
from idaes.models.unit_models.gibbs_reactor import GibbsReactor, GibbsReactorScaler
from idaes.models.properties.activity_coeff_models.methane_combustion_ideal import (
    MethaneParameterBlock as MethaneCombustionParameterBlock,
)
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.scaling import (
    jacobian_cond,
)
from idaes.core.util import from_json, StoreSpec
from idaes.core.scaling import CustomScalerBase, get_scaling_factor, set_scaling_factor


# Get solution json from scaling tests
FILENAME = "gibbs_solution.json"
local_path = os.path.dirname(os.path.realpath(__file__))
fname = os.path.join(local_path, "..", "..", "..", "core", "scaling", "tests", FILENAME)


# -----------------------------------------------------------------------------
@pytest.fixture
def test_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = GibbsReactor(
        property_package=m.fs.properties,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    return m


class DummyScaler:
    def variable_scaling_routine(self, model, overwrite):
        model._dummy_scaler_test = overwrite

    def constraint_scaling_routine(self, model, overwrite):
        model._dummy_scaler_test = overwrite


@pytest.mark.unit
class TestVariableScaling:

    def test_variable_scaling_no_input(self, test_model):
        scaler = GibbsReactorScaler()

        scaler.variable_scaling_routine(test_model.fs.unit)

        for v in test_model.fs.unit.lagrange_mult.values():
            assert test_model.fs.unit.scaling_factor[v] == pytest.approx(
                1 / (8.314 * 500), rel=1e-4
            )

        for v in test_model.fs.unit.control_volume.heat.values():
            assert test_model.fs.unit.control_volume.scaling_factor[v] == pytest.approx(
                1e-6, rel=1e-4
            )

        for v in test_model.fs.unit.control_volume.deltaP.values():
            assert test_model.fs.unit.control_volume.scaling_factor[v] == pytest.approx(
                1e-3, rel=1e-4
            )

    def test_variable_scaling_no_heat_deltaP(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = GibbsReactor(
            property_package=m.fs.properties,
            has_heat_transfer=False,
            has_pressure_change=False,
        )

        scaler = GibbsReactorScaler()

        scaler.variable_scaling_routine(m.fs.unit)

        for v in m.fs.unit.lagrange_mult.values():
            assert m.fs.unit.scaling_factor[v] == pytest.approx(
                1 / (8.314 * 500), rel=1e-4
            )

    def test_variable_scaling_inlet_state(self, test_model):
        prop_in = test_model.fs.unit.control_volume.properties_in[0]
        sfx = prop_in.scaling_factor = Suffix(direction=Suffix.EXPORT)
        sfx[prop_in.temperature] = 1e-2
        sfx[prop_in.pressure] = 1e-5
        for j in prop_in.flow_mol_phase_comp.values():
            sfx[j] = 1e-2

        scaler = GibbsReactorScaler()

        scaler.variable_scaling_routine(test_model.fs.unit)

        # Outlet properties should now have scaling factors
        prop_out = test_model.fs.unit.control_volume.properties_out[0]
        assert prop_out.scaling_factor[prop_out.temperature] == 1e-2
        assert prop_out.scaling_factor[prop_out.pressure] == 1e-5
        for j in prop_out.flow_mol_phase_comp.values():
            prop_out.scaling_factor[j] == 1e-2

        for v in test_model.fs.unit.lagrange_mult.values():
            assert test_model.fs.unit.scaling_factor[v] == pytest.approx(
                1 / (8.314 * 100), rel=1e-4
            )

        for v in test_model.fs.unit.control_volume.heat.values():
            assert test_model.fs.unit.control_volume.scaling_factor[v] == pytest.approx(
                1e-6, rel=1e-4
            )

        for v in test_model.fs.unit.control_volume.deltaP.values():
            assert test_model.fs.unit.control_volume.scaling_factor[v] == pytest.approx(
                1e-3, rel=1e-4
            )

    def test_variable_scaling_submodel_scalers(self, test_model):
        scaler = GibbsReactorScaler()

        scaler_map = ComponentMap()
        scaler_map[test_model.fs.unit.control_volume.properties_in] = DummyScaler()
        scaler_map[test_model.fs.unit.control_volume.properties_out] = DummyScaler()

        scaler.variable_scaling_routine(
            test_model.fs.unit,
            submodel_scalers=scaler_map,
        )

        # Check to see if testing attribute was created correctly
        assert not test_model.fs.unit.control_volume.properties_in[0]._dummy_scaler_test
        assert not test_model.fs.unit.control_volume.properties_out[
            0
        ]._dummy_scaler_test


@pytest.mark.unit
class TestConstraintScaling:

    def test_constraint_scaling_no_inputs(self, test_model):
        scaler = GibbsReactorScaler()

        scaler.constraint_scaling_routine(test_model.fs.unit)

        sfx = test_model.fs.unit.control_volume.scaling_factor

        assert sfx[
            test_model.fs.unit.control_volume.element_balances[0.0, "H"]
        ] == pytest.approx(0.05, rel=1e-5)
        assert sfx[
            test_model.fs.unit.control_volume.element_balances[0.0, "He"]
        ] == pytest.approx(0.0357143, rel=1e-5)
        assert sfx[
            test_model.fs.unit.control_volume.element_balances[0.0, "Li"]
        ] == pytest.approx(0.0277778, rel=1e-5)
        assert sfx[
            test_model.fs.unit.control_volume.enthalpy_balances[0.0]
        ] == pytest.approx(0.25, rel=1e-5)
        assert sfx[
            test_model.fs.unit.control_volume.pressure_balance[0.0]
        ] == pytest.approx(5e-6, rel=1e-5)

        for k, v in test_model.fs.unit.gibbs_minimization.items():
            if k[2] == "c1":
                assert test_model.fs.unit.scaling_factor[v] == pytest.approx(
                    1.53846e-3, rel=1e-5
                )
            else:
                assert test_model.fs.unit.scaling_factor[v] == pytest.approx(
                    6.45161e-4, rel=1e-5
                )

    def test_constraint_scaling_inerts(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = GibbsReactor(
            property_package=m.fs.properties,
            has_heat_transfer=True,
            has_pressure_change=True,
            inert_species=["c1"],
        )

        scaler = GibbsReactorScaler()

        scaler.constraint_scaling_routine(m.fs.unit)

        sfx = m.fs.unit.control_volume.scaling_factor

        assert sfx[
            m.fs.unit.control_volume.element_balances[0.0, "H"]
        ] == pytest.approx(0.05, rel=1e-5)
        assert sfx[
            m.fs.unit.control_volume.element_balances[0.0, "He"]
        ] == pytest.approx(0.0357143, rel=1e-5)
        assert sfx[
            m.fs.unit.control_volume.element_balances[0.0, "Li"]
        ] == pytest.approx(0.0277778, rel=1e-5)
        assert sfx[m.fs.unit.control_volume.enthalpy_balances[0.0]] == pytest.approx(
            0.25, rel=1e-5
        )
        assert sfx[m.fs.unit.control_volume.pressure_balance[0.0]] == pytest.approx(
            5e-6, rel=1e-5
        )

        for k, v in m.fs.unit.gibbs_minimization.items():
            assert m.fs.unit.scaling_factor[v] == pytest.approx(6.45161e-4, rel=1e-5)

        for k, v in m.fs.unit.inert_species_balance.items():
            assert m.fs.unit.scaling_factor[v] == pytest.approx(0.5, rel=1e-5)

    def test_constraint_scaling_submodel_scalers(self, test_model):
        scaler = GibbsReactorScaler()

        scaler_map = ComponentMap()
        scaler_map[test_model.fs.unit.control_volume.properties_in] = DummyScaler()
        scaler_map[test_model.fs.unit.control_volume.properties_out] = DummyScaler()

        scaler.constraint_scaling_routine(
            test_model.fs.unit,
            submodel_scalers=scaler_map,
        )

        # Check to see if testing attribute was created correctly
        assert not test_model.fs.unit.control_volume.properties_in[0]._dummy_scaler_test
        assert not test_model.fs.unit.control_volume.properties_out[
            0
        ]._dummy_scaler_test


# -----------------------------------------------------------------------------
class SMScaler(CustomScalerBase):
    def variable_scaling_routine(self, model, overwrite):
        pass

    def constraint_scaling_routine(self, model, overwrite):
        for c in model.component_data_objects(ctype=Constraint, descend_into=True):
            self.scale_constraint_by_nominal_value(
                c, scheme="inverse_sum", overwrite=overwrite
            )


# TODO: Turn this into a testing harness?
@pytest.mark.integration
class TestMethaneScaling(object):
    @pytest.fixture
    def methane(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)

        model.fs.properties = MethaneCombustionParameterBlock()

        model.fs.unit = GibbsReactor(
            property_package=model.fs.properties,
            has_heat_transfer=True,
            has_pressure_change=True,
        )

        model.fs.unit.inlet.flow_mol[0].fix(230.0)
        model.fs.unit.inlet.mole_frac_comp[0, "H2"].fix(0.0435)
        model.fs.unit.inlet.mole_frac_comp[0, "N2"].fix(0.6522)
        model.fs.unit.inlet.mole_frac_comp[0, "O2"].fix(0.1739)
        model.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(1e-5)
        model.fs.unit.inlet.mole_frac_comp[0, "CH4"].fix(0.1304)
        model.fs.unit.inlet.mole_frac_comp[0, "CO"].fix(1e-5)
        model.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(1e-5)
        model.fs.unit.inlet.mole_frac_comp[0, "NH3"].fix(1e-5)
        model.fs.unit.inlet.temperature[0].fix(1500.0)
        model.fs.unit.inlet.pressure[0].fix(101325.0)

        model.fs.unit.outlet.temperature[0].fix(2844.38)
        model.fs.unit.deltaP.fix(0)

        # Set imperfect scaling factors for all variables, representing an initial "best-guess"
        # Feed states are known exactly - set scaling based on these
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].flow_mol, 1 / 230
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].flow_mol_phase, 1 / 230
        )  # Only 1 phase, so we "know" this
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["H2"],
            1 / 0.0435,
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["N2"],
            1 / 0.6522,
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["O2"],
            1 / 0.1739,
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CO2"], 1e5
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CH4"],
            1 / 0.1304,
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CO"], 1e5
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["H2O"], 1e5
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["NH3"], 1e5
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].temperature, 1 / 1500
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_in[0.0].pressure, 1e-5
        )
        # Assume user does not know anything about enthalpy

        # Best guesses for unit model and outlet state conditions
        set_scaling_factor(model.fs.unit.control_volume.heat[0.0], 1e-6)

        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].flow_mol, 1e-2
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].flow_mol_phase, 1e-2
        )  # Only 1 phase, so we "know" this
        # N2 is inert, so will be order 0.1, assume CH4 and H2 are near-totally consumed, assume most O2 consumed
        # Assume moderate amounts of CO2 and H2O, small amounts of CO, trace NH3 NH3
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2"], 1e4
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["N2"], 10
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["O2"], 1e2
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO2"], 10
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CH4"], 1e4
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO"], 1e3
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2O"], 10
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["NH3"], 1e4
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].temperature, 1e-3
        )
        set_scaling_factor(
            model.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5
        )

        from_json(model, fname=fname, wts=StoreSpec.value())

        return model

    def test_variable_scaling_only(self, methane):
        unscaled = jacobian_cond(methane, scaled=False)

        scaler_map = ComponentMap()
        scaler_map[methane.fs.unit.control_volume.properties_in] = SMScaler()
        scaler_map[methane.fs.unit.control_volume.properties_out] = SMScaler()

        scaler = GibbsReactorScaler()
        scaler.variable_scaling_routine(
            methane.fs.unit,
            submodel_scalers=scaler_map,
        )

        scaled = jacobian_cond(methane, scaled=True)

        count = 0
        for c in methane.component_data_objects(ctype=[Var], descend_into=True):
            sf = get_scaling_factor(c)
            if sf is None:
                count += 1
        assert count == 52

        assert scaled < unscaled
        assert scaled == pytest.approx(8.908989e16, rel=1e-5)

    def test_constraint_scaling_only(self, methane):
        unscaled = jacobian_cond(methane, scaled=False)

        scaler_map = ComponentMap()
        scaler_map[methane.fs.unit.control_volume.properties_in] = SMScaler()
        scaler_map[methane.fs.unit.control_volume.properties_out] = SMScaler()

        scaler = GibbsReactorScaler()
        scaler.constraint_scaling_routine(
            methane.fs.unit,
            submodel_scalers=scaler_map,
        )

        scaled = jacobian_cond(methane, scaled=True)

        count = 0
        for c in methane.component_data_objects(ctype=[Constraint], descend_into=True):
            sf = get_scaling_factor(c)
            if sf is None:
                count += 1
        assert count == 0

        assert scaled < unscaled
        assert scaled == pytest.approx(9.316e15, rel=1e-2)

    def test_full_scaling(self, methane):
        unscaled = jacobian_cond(methane, scaled=False)

        scaler_map = ComponentMap()
        scaler_map[methane.fs.unit.control_volume.properties_in] = SMScaler()
        scaler_map[methane.fs.unit.control_volume.properties_out] = SMScaler()

        scaler = GibbsReactorScaler()
        scaler.scale_model(
            methane.fs.unit,
            submodel_scalers=scaler_map,
        )

        scaled = jacobian_cond(methane, scaled=True)

        assert scaled < unscaled
        assert scaled == pytest.approx(7.653e15, rel=1e-2)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import inspect
import pytest
from types import MethodType

from pyomo.environ import ConcreteModel, Block, Set, units, value, Var
from pyomo.common.config import ConfigBlock, ConfigValue
from idaes.core import (
    ControlVolumeBlockData,
    CONFIG_Template,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    declare_process_block_class,
    FlowsheetBlockData,
    UnitModelBlockData,
    useDefault,
    PhysicalParameterBlock,
    ReactionParameterBlock,
)
from idaes.core.util.exceptions import (
    ConfigurationError,
    DynamicError,
    PropertyPackageError,
)


# -----------------------------------------------------------------------------
# Test Enumerators for balance type options
@pytest.mark.unit
def test_material_balance_type():
    assert len(MaterialBalanceType) == 6

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        MaterialBalanceType.foo  # pylint: disable=no-member


@pytest.mark.unit
def test_energy_balance_type():
    assert len(EnergyBalanceType) == 6

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        EnergyBalanceType.foo  # pylint: disable=no-member


@pytest.mark.unit
def test_momentum_balance_type():
    assert len(MomentumBalanceType) == 5

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        MomentumBalanceType.foo  # pylint: disable=no-member


@pytest.mark.unit
def testflow_direction():
    assert len(FlowDirection) == 2

    # Test that error is raised when given non-member
    with pytest.raises(AttributeError):
        FlowDirection.foo  # pylint: disable=no-member


# -----------------------------------------------------------------------------
# Test CONFIG_Template
@pytest.mark.unit
def test_CONFIG_Template():
    c = CONFIG_Template()

    assert len(c) == 18

    for i in c:
        if i == "dynamic":
            assert c[i] == useDefault
        elif i == "material_balance_type":
            assert c[i] == MaterialBalanceType.componentPhase
        elif i == "energy_balance_type":
            assert c[i] == EnergyBalanceType.enthalpyTotal
        elif i == "momentum_balance_type":
            assert c[i] == MomentumBalanceType.pressureTotal
        elif i == "property_package":
            assert c[i] == useDefault
        elif i == "reaction_package":
            assert c[i] is None
        elif i in ["property_package_args", "reaction_package_args"]:
            assert isinstance(c[i], ConfigBlock)
            assert len(c[i]) == 0
        else:
            assert c[i] is False


@pytest.mark.unit
def test_CONFIG_Template_validation_general():
    # No config argument takes a string, float/int or list
    c = CONFIG_Template()

    for i in c:
        with pytest.raises(ValueError):
            c[i] = "foo"
        with pytest.raises(ValueError):
            c[i] = 10.0
        with pytest.raises(ValueError):
            c[i] = [1, 2]


@pytest.mark.unit
def test_CONFIG_Template_true_false():
    # Check arguments that accept True/False as values
    c = CONFIG_Template()

    for i in c:
        if i not in [
            "material_balance_type",
            "energy_balance_type",
            "momentum_balance_type",
            "property_package",
            "reaction_package",
            "property_package_args",
            "reaction_package_args",
        ]:
            c[i] = True
            c[i] = False


@pytest.mark.unit
def test_CONFIG_Template_material_balance_type():
    c = CONFIG_Template()

    for i in MaterialBalanceType:
        c["material_balance_type"] = i


@pytest.mark.unit
def test_CONFIG_Template_energy_balance_type():
    c = CONFIG_Template()

    for i in EnergyBalanceType:
        c["energy_balance_type"] = i


@pytest.mark.unit
def test_CONFIG_Template_momentum_balance_type():
    c = CONFIG_Template()

    for i in MomentumBalanceType:
        c["momentum_balance_type"] = i


# -----------------------------------------------------------------------------
# Mockup classes for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class _UnitData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare("property_package", ConfigValue(default=None))
    CONFIG.declare("property_package_args", ConfigValue(default={}))

    def build(self):
        super(_UnitData, self).build()


# -----------------------------------------------------------------------------
# Testing ControlVolumeBlockData
@declare_process_block_class("CVFrame")
class CVFrameData(ControlVolumeBlockData):
    def build(self):
        super(ControlVolumeBlockData, self).build()


@pytest.mark.unit
def test_config_block():
    cv = CVFrame(concrete=True)

    assert len(cv.config) == 7
    assert cv.config.dynamic == useDefault
    assert cv.config.has_holdup is useDefault
    assert cv.config.property_package == useDefault
    assert isinstance(cv.config.property_package_args, ConfigBlock)
    assert len(cv.config.property_package_args) == 0
    assert cv.config.reaction_package is None
    assert isinstance(cv.config.reaction_package_args, ConfigBlock)
    assert len(cv.config.reaction_package_args) == 0
    assert cv.config.auto_construct is False


# -----------------------------------------------------------------------------
# Test _setup_dynamics
@pytest.mark.unit
def test_setup_dynamics_use_parent_value():
    # Test that dynamic = None works correctly
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.u = Unit(dynamic=False)
    m.fs.u.cv = CVFrame()

    m.fs.u.cv._setup_dynamics()

    assert m.fs.u.cv.config.dynamic is False
    assert m.fs.u.cv.config.has_holdup is False


@pytest.mark.unit
def test_setup_dynamics_use_parent_value_fail_no_dynamic():
    # Test that default falls back to flowsheet
    fs = Flowsheet(dynamic=False, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    fs.b.cv = CVFrame()
    fs.b.cv._setup_dynamics()

    assert fs.b.cv.config.dynamic is False


@pytest.mark.unit
def test_setup_dynamics_dynamic_in_ss():
    # Test that dynamic = None works correctly
    fs = Flowsheet(dynamic=False, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    # Add a time attribute to make sure the correct failure triggers
    fs.b.time_ref = Set(initialize=[0])

    fs.b.cv = CVFrame(dynamic=True, has_holdup=True)

    # _setup_dynamics should return DynamicError
    with pytest.raises(DynamicError):
        fs.b.cv._setup_dynamics()


@pytest.mark.unit
def test_setup_dynamics_dynamic_holdup_inconsistent():
    # Test that dynamic = None works correctly
    fs = Flowsheet(dynamic=True, time_units=units.s, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    # Add a time attribute to make sure the correct failure triggers
    fs.b.time_ref = Set(initialize=[0])

    fs.b.cv = CVFrame(dynamic=True, has_holdup=False)

    # _setup_dynamics should return ConfigurationError
    with pytest.raises(ConfigurationError):
        fs.b.cv._setup_dynamics()


# -----------------------------------------------------------------------------
# Test _get_property_package
@declare_process_block_class("PropertyParameterBlock")
class _PropertyParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PropertyParameterBlock, self).build()

        frm = inspect.stack()[1]
        self._package_module = inspect.getmodule(frm[0])

        self.phase_list = Set(initialize=["p1", "p2"])
        self.component_list = Set(initialize=["c1", "c2"])


@pytest.mark.unit
def test_get_property_package_set():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock()
    m.cv = CVFrame(property_package=m.pp)
    m.cv._get_property_package()


@pytest.mark.unit
def test_get_property_package_default_args():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock(default_arguments={"test": "foo"})
    m.cv = CVFrame(property_package=m.pp)
    m.cv._get_property_package()

    assert m.cv.config.property_package_args["test"] == "foo"


@pytest.mark.unit
def test_get_reaction_package_module_combine_args():
    # Test that local and default args combine correctly
    m = ConcreteModel()
    m.pp = PropertyParameterBlock(default_arguments={"test1": "foo", "test2": "bar"})
    m.cv = CVFrame(
        property_package=m.pp, property_package_args={"test2": "baz", "test3": "bar"}
    )
    m.cv._get_property_package()

    assert m.cv.config.property_package_args["test1"] == "foo"
    assert m.cv.config.property_package_args["test2"] == "baz"
    assert m.cv.config.property_package_args["test3"] == "bar"


# -----------------------------------------------------------------------------
# Test _get_default_prop_pack
@pytest.mark.unit
def test_get_default_prop_pack_works():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.pp

    m.fs.cv = CVFrame()
    assert m.fs.cv._get_default_prop_pack() == m.fs.pp


# TODO : should test more failure modes
@pytest.mark.unit
def test_get_default_prop_pack_no_default():
    m = ConcreteModel()
    m.fs = Flowsheet()

    m.fs.cv = CVFrame()
    with pytest.raises(ConfigurationError):
        m.fs.cv._get_default_prop_pack()


@pytest.mark.unit
def test_get_property_package_call_to_get_default_prop_pack():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.pp

    m.fs.cv = CVFrame()
    m.fs.cv._get_property_package()
    assert m.fs.cv.config.property_package == m.fs.pp


# -----------------------------------------------------------------------------
# Test _get_indexing_sets
@pytest.mark.unit
def test_get_indexing_sets_missing_phase_list():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock()
    m.pp.del_component(m.pp.phase_list)
    m.cv = CVFrame(property_package=m.pp)
    m.cv._get_property_package()

    with pytest.raises(PropertyPackageError):
        m.cv._get_indexing_sets()


@pytest.mark.unit
def test_get_indexing_sets_missing_component_list():
    m = ConcreteModel()
    m.pp = PropertyParameterBlock()
    m.pp.del_component(m.pp.component_list)
    m.cv = CVFrame(property_package=m.pp)
    m.cv._get_property_package()

    with pytest.raises(PropertyPackageError):
        m.cv._get_indexing_sets()


# -----------------------------------------------------------------------------
# Test _get_reaction_package
@pytest.mark.unit
def test_get_reaction_package_none():
    m = ConcreteModel()
    m.r = CVFrame()

    m.r._get_reaction_package()

    assert hasattr(m.r, "reaction_module") is False


@declare_process_block_class("ReactionParameterTestBlock")
class _ReactionParameterBlock(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

        frm = inspect.stack()[1]
        self._package_module = inspect.getmodule(frm[0])


@pytest.mark.unit
def test_get_reaction_package_module():
    m = ConcreteModel()
    m.rp = ReactionParameterTestBlock(default_arguments={"test": "foo"})
    m.cv = CVFrame(reaction_package=m.rp)

    m.cv._get_reaction_package()

    assert m.cv.config.reaction_package == m.rp
    assert m.cv.config.reaction_package_args["test"] == "foo"


@pytest.mark.unit
def test_get_reaction_package_module_default_args():
    # Test that local and default args combine correctly
    m = ConcreteModel()
    m.rp = ReactionParameterTestBlock(
        default_arguments={"test1": "foo", "test2": "bar"}
    )
    m.cv = CVFrame(
        reaction_package=m.rp, reaction_package_args={"test2": "baz", "test3": "bar"}
    )

    m.cv._get_reaction_package()

    assert m.cv.config.reaction_package_args["test1"] == "foo"
    assert m.cv.config.reaction_package_args["test2"] == "baz"
    assert m.cv.config.reaction_package_args["test3"] == "bar"


# -----------------------------------------------------------------------------
# Test build and auto_construct methods
@pytest.mark.unit
def test_build():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.cv = CVFrame(property_package=m.fs.pp)

    super(CVFrameData, m.fs.cv).build()


@pytest.mark.unit
def test_add_geometry():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.cv = CVFrame()

    with pytest.raises(NotImplementedError):
        m.fs.cv.add_geometry()


@pytest.mark.unit
def test_auto_construct():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PropertyParameterBlock()
    m.fs.cv = CVFrame(property_package=m.fs.pp, auto_construct=True)

    with pytest.raises(NotImplementedError):
        super(CVFrameData, m.fs.cv).build()


# -----------------------------------------------------------------------------
# Test NotImplementedErrors for all property and balance type methods
@pytest.mark.unit
def test_add_state_blocks():
    m = ConcreteModel()
    m.cv = CVFrame()

    with pytest.raises(NotImplementedError):
        m.cv.add_state_blocks()


@pytest.mark.unit
def test_add_reaction_blocks():
    m = ConcreteModel()
    m.cv = CVFrame()

    with pytest.raises(NotImplementedError):
        m.cv.add_reaction_blocks()


@pytest.mark.unit
def test_add_material_balances():
    m = ConcreteModel()
    m.cv = CVFrame()

    for t in MaterialBalanceType:
        if t == MaterialBalanceType.none:
            assert m.cv.add_material_balances(t) is None
        elif t == MaterialBalanceType.useDefault:
            with pytest.raises(ConfigurationError):
                m.cv.add_material_balances(t)
        else:
            with pytest.raises(NotImplementedError):
                m.cv.add_material_balances(t)


@pytest.mark.unit
def test_add_energy_balances():
    m = ConcreteModel()
    m.cv = CVFrame()

    for t in EnergyBalanceType:
        if t == EnergyBalanceType.none:
            assert m.cv.add_energy_balances(t) is None
        elif t == EnergyBalanceType.useDefault:
            with pytest.raises(ConfigurationError):
                m.cv.add_energy_balances(t)
        else:
            with pytest.raises(NotImplementedError):
                m.cv.add_energy_balances(t)


@pytest.mark.unit
def test_add_momentum_balances():
    m = ConcreteModel()
    m.cv = CVFrame()

    for t in MomentumBalanceType:
        if t == MomentumBalanceType.none:
            assert m.cv.add_momentum_balances(t) is None
        else:
            with pytest.raises(NotImplementedError):
                m.cv.add_momentum_balances(t)


@pytest.mark.unit
def test_estimate_state_var_generic():
    m = ConcreteModel()
    m.cv = CVFrame()

    m.cv.b1 = Block()
    m.cv.b1.state_var = Var(initialize=16)

    m.cv.b2 = Block()
    m.cv.b2.state_var = Var(initialize=17)

    # Has value and always_estimate False, expect no effect
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.state_var, index=None, always_estimate=False
    )
    assert value(m.cv.b2.state_var) == 17

    # Has value but always_estimate True, expect value equals b1.state_var
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.state_var, index=None, always_estimate=True
    )
    assert value(m.cv.b2.state_var) == 16

    # Fixed and always_estimate True, expect value to remain fixed
    m.cv.b2.state_var.fix(24)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.state_var, index=None, always_estimate=True
    )
    assert value(m.cv.b2.state_var) == 24


@pytest.mark.unit
def test_estimate_state_var_temperature():
    m = ConcreteModel()
    m.cv = CVFrame()
    m.cv.myset = Set(initialize=["a", "b"])

    m.cv.b1 = Block()
    m.cv.b1.state_var = Var(initialize=16)

    m.cv.b2 = Block()
    m.cv.b2.temperature = Var(initialize=17)

    # Has value and always_estimate False, expect no effect
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.temperature, index="a", always_estimate=False
    )
    assert value(m.cv.b2.temperature) == 17

    # Has value but always_estimate True, expect to estimate
    # No deltaT defined, should have value of b1.state_var
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.temperature, index="a", always_estimate=True
    )
    assert value(m.cv.b2.temperature) == 16

    # deltaT defined but not fixed, should keep value of b1.state_var
    m.cv.b2.temperature.set_value(17)
    m.cv.deltaT = Var(m.cv.myset)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.temperature, index="a", always_estimate=True
    )
    assert value(m.cv.b2.temperature) == 16

    # deltaT defined and now fixed, should keep value of b1.state_var+deltaT
    m.cv.b2.temperature.set_value(17)
    m.cv.deltaT["a"].fix(10)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.temperature, index="a", always_estimate=True
    )
    assert value(m.cv.b2.temperature) == 26

    # Fixed and always_estimate True, expect value to remain fixed
    m.cv.b2.temperature.fix(24)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.temperature, index="a", always_estimate=True
    )
    assert value(m.cv.b2.temperature) == 24


@pytest.mark.unit
def test_estimate_state_var_pressure():
    m = ConcreteModel()
    m.cv = CVFrame()
    m.cv.myset = Set(initialize=["a", "b"])

    m.cv.b1 = Block()
    m.cv.b1.state_var = Var(initialize=16)

    m.cv.b2 = Block()
    m.cv.b2.pressure = Var(initialize=17)

    # Has value and always_estimate False, expect no effect
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.pressure, index="a", always_estimate=False
    )
    assert value(m.cv.b2.pressure) == 17

    # Has value but always_estimate True, expect to estimate
    # No deltaP defined, should have value of b1.state_var
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.pressure, index="a", always_estimate=True
    )
    assert value(m.cv.b2.pressure) == 16

    # deltaP defined but not fixed, should keep value of b1.state_var
    m.cv.b2.pressure.set_value(17)
    m.cv.deltaP = Var(m.cv.myset)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.pressure, index="a", always_estimate=True
    )
    assert value(m.cv.b2.pressure) == 16

    # deltaP defined and now fixed, should keep value of b1.state_var+deltaT
    m.cv.b2.pressure.set_value(17)
    m.cv.deltaP["a"].fix(10)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.pressure, index="a", always_estimate=True
    )
    assert value(m.cv.b2.pressure) == 26

    # Fixed and always_estimate True, expect value to remain fixed
    m.cv.b2.pressure.fix(24)
    m.cv._estimate_state_var(
        m.cv.b1.state_var, m.cv.b2.pressure, index="a", always_estimate=True
    )
    assert value(m.cv.b2.pressure) == 24


@pytest.mark.unit
def test_estimate_next_state_no_values():
    m = ConcreteModel()
    m.cv = CVFrame()
    m.cv.myset = Set(initialize=["a", "b"])

    m.cv.b1 = Block(m.cv.myset)
    m.cv.b1["a"].state_var = Var(m.cv.myset, initialize={"a": 12, "b": 13})
    m.cv.b1["a"].temperature = Var(initialize=14)
    m.cv.b1["a"].pressure = Var(initialize=15)
    m.cv.b1["a"].energy_internal = Var(initialize=16)
    m.cv.b1["b"].state_var = Var(m.cv.myset, initialize={"a": 22, "b": 23})
    m.cv.b1["b"].temperature = Var(initialize=24)
    m.cv.b1["b"].pressure = Var(initialize=25)
    m.cv.b1["b"].energy_internal = Var(initialize=26)

    m.cv.b2 = Block(m.cv.myset)
    m.cv.b2["a"].state_var = Var(m.cv.myset)
    m.cv.b2["a"].temperature = Var()
    m.cv.b2["a"].pressure = Var()
    m.cv.b2["a"].energy_internal = Var()
    m.cv.b2["b"].state_var = Var(m.cv.myset)
    m.cv.b2["b"].temperature = Var()
    m.cv.b2["b"].pressure = Var()
    m.cv.b2["b"].energy_internal = Var()

    def define_state_vars(blk):
        return {
            "foo": blk.state_var,
            "bar": blk.temperature,
            "baz": blk.pressure,
            "baz2": blk.energy_internal,
        }

    m.cv.b2["a"].define_state_vars = MethodType(define_state_vars, m.cv.b2["a"])

    m.cv._estimate_next_state(m.cv.b1["a"], m.cv.b2["a"], "a", False)

    assert value(m.cv.b2["a"].state_var["a"]) == 12
    assert value(m.cv.b2["a"].state_var["b"]) == 13
    assert value(m.cv.b2["a"].temperature) == 14
    assert value(m.cv.b2["a"].pressure) == 15
    assert value(m.cv.b2["a"].energy_internal) == 16

    assert m.cv.b2["b"].state_var["a"]._value is None
    assert m.cv.b2["b"].state_var["b"]._value is None
    assert m.cv.b2["b"].temperature._value is None
    assert m.cv.b2["b"].pressure._value is None
    assert m.cv.b2["b"].energy_internal._value is None


@pytest.mark.unit
def test_estimate_next_state_always_estimate_false():
    m = ConcreteModel()
    m.cv = CVFrame()
    m.cv.myset = Set(initialize=["a", "b"])

    m.cv.b1 = Block(m.cv.myset)
    m.cv.b1["a"].state_var = Var(m.cv.myset, initialize={"a": 12, "b": 13})
    m.cv.b1["a"].temperature = Var(initialize=14)
    m.cv.b1["a"].pressure = Var(initialize=15)
    m.cv.b1["a"].energy_internal = Var(initialize=16)
    m.cv.b1["b"].state_var = Var(m.cv.myset, initialize={"a": 22, "b": 23})
    m.cv.b1["b"].temperature = Var(initialize=24)
    m.cv.b1["b"].pressure = Var(initialize=25)
    m.cv.b1["b"].energy_internal = Var(initialize=26)

    m.cv.b2 = Block(m.cv.myset)
    m.cv.b2["a"].state_var = Var(m.cv.myset, initialize=1)
    m.cv.b2["a"].temperature = Var(initialize=1)
    m.cv.b2["a"].pressure = Var(initialize=1)
    m.cv.b2["a"].energy_internal = Var(initialize=1)
    m.cv.b2["b"].state_var = Var(m.cv.myset, initialize=1)
    m.cv.b2["b"].temperature = Var(initialize=1)
    m.cv.b2["b"].pressure = Var(initialize=1)
    m.cv.b2["b"].energy_internal = Var(initialize=1)

    def define_state_vars(blk):
        return {
            "foo": blk.state_var,
            "bar": blk.temperature,
            "baz": blk.pressure,
            "baz2": blk.energy_internal,
        }

    m.cv.b2["a"].define_state_vars = MethodType(define_state_vars, m.cv.b2["a"])

    m.cv._estimate_next_state(m.cv.b1["a"], m.cv.b2["a"], "a", False)

    assert value(m.cv.b2["a"].state_var["a"]) == 1
    assert value(m.cv.b2["a"].state_var["b"]) == 1
    assert value(m.cv.b2["a"].temperature) == 1
    assert value(m.cv.b2["a"].pressure) == 1
    assert value(m.cv.b2["a"].energy_internal) == 1

    assert value(m.cv.b2["b"].state_var["a"]) == 1
    assert value(m.cv.b2["b"].state_var["b"]) == 1
    assert value(m.cv.b2["b"].temperature) == 1
    assert value(m.cv.b2["b"].pressure) == 1
    assert value(m.cv.b2["b"].energy_internal) == 1


@pytest.mark.unit
def test_estimate_next_state_always_estimate_true():
    m = ConcreteModel()
    m.cv = CVFrame()
    m.cv.myset = Set(initialize=["a", "b"])

    m.cv.b1 = Block(m.cv.myset)
    m.cv.b1["a"].state_var = Var(m.cv.myset, initialize={"a": 12, "b": 13})
    m.cv.b1["a"].temperature = Var(initialize=14)
    m.cv.b1["a"].pressure = Var(initialize=15)
    m.cv.b1["a"].energy_internal = Var(initialize=16)
    m.cv.b1["b"].state_var = Var(m.cv.myset, initialize={"a": 22, "b": 23})
    m.cv.b1["b"].temperature = Var(initialize=24)
    m.cv.b1["b"].pressure = Var(initialize=25)
    m.cv.b1["b"].energy_internal = Var(initialize=26)

    m.cv.b2 = Block(m.cv.myset)
    m.cv.b2["a"].state_var = Var(m.cv.myset, initialize=1)
    m.cv.b2["a"].temperature = Var(initialize=1)
    m.cv.b2["a"].pressure = Var(initialize=1)
    m.cv.b2["a"].energy_internal = Var(initialize=1)
    m.cv.b2["b"].state_var = Var(m.cv.myset, initialize=1)
    m.cv.b2["b"].temperature = Var(initialize=1)
    m.cv.b2["b"].pressure = Var(initialize=1)
    m.cv.b2["b"].energy_internal = Var(initialize=1)

    def define_state_vars(blk):
        return {
            "foo": blk.state_var,
            "bar": blk.temperature,
            "baz": blk.pressure,
            "baz2": blk.energy_internal,
        }

    m.cv.b2["b"].define_state_vars = MethodType(define_state_vars, m.cv.b2["b"])

    m.cv._estimate_next_state(m.cv.b1["b"], m.cv.b2["b"], "b", True)

    assert value(m.cv.b2["a"].state_var["a"]) == 1
    assert value(m.cv.b2["a"].state_var["b"]) == 1
    assert value(m.cv.b2["a"].temperature) == 1
    assert value(m.cv.b2["a"].pressure) == 1
    assert value(m.cv.b2["a"].energy_internal) == 1

    assert value(m.cv.b2["b"].state_var["a"]) == 22
    assert value(m.cv.b2["b"].state_var["b"]) == 23
    assert value(m.cv.b2["b"].temperature) == 24
    assert value(m.cv.b2["b"].pressure) == 25
    assert value(m.cv.b2["b"].energy_internal) == 26
